import functools as _functools
import importlib
import logging
from typing import TYPE_CHECKING
from typing import Dict as _Dict
from typing import List as _List
from typing import Union as _Union

import torch.nn as _nn

from tamm import _warnings as _tamm_warnings
from tamm._plugin import discover_and_retrieve_plugins
from tamm.model_repo.file import FileModelRepo
from tamm.model_repo.model_repo import BaseModelRepo
from tamm.model_repo.uri_handler import URIHandlerModelRepo

if TYPE_CHECKING:
    from tamm.layers.common import LayerBuilder
    from tamm.models.common import ModelConfig
logger = logging.getLogger(__name__)


@_functools.lru_cache(maxsize=1)
def get_model_repo_lazy(*args, **kwargs) -> "BaseModelRepo":
    """
    Creates model repo lazily by calling :py:func:`tamm.model_repo.utils.get_model_repo_eager`
    whenever it is first needed.

    .. note::
        This reduces side effects when ``import tamm`` is executed.
    """

    return get_model_repo_eager(*args, **kwargs)


def create_model(model_name: str, *args, **kwargs) -> "_nn.Module":
    """
    Create a model instance for the specified model name.

    Args:
        model_name (:obj:`str`): The name of the model to be created.
        *args: Additional positional arguments for model creation.
        **kwargs: Additional keyword arguments for model creation.

    Returns:
        :obj:`_nn.Module`: Created model instance.
    """
    return get_model_repo_lazy().create_model(model_name, *args, **kwargs)


def create_adapted_model(model_name: str, *args, **kwargs) -> "_nn.Module":
    """
    Similar to `create_model()` but accepts some additional LoRA options in
        ``**kwargs`` for backward compatibility.

    .. admonition:: Deprecation warning
        :class: warning

        This function is deprecated.  Please use `tamm.create_model` instead.

    Args:
        model_name (:obj:`str`): The name or identifier of the model to be created.
        *args: Additional positional arguments for model creation.
        **kwargs: Additional keyword arguments for model creation.

    Returns:
        :obj:`_nn.Module`: The created adapted model instance.
    """
    return get_model_repo_lazy().create_adapted_model(model_name, *args, **kwargs)


def create_model_builder(model_name: str, *args, **kwargs) -> "LayerBuilder":
    """
    Create a model builder for the specified model.

    Args:
        model_name (:obj:`str`): The name of the model for which the builder is to be
            created.
        *args: Additional positional arguments for builder creation.
        **kwargs: Additional keyword arguments for builder creation.

    Returns:
        :obj:`layer_common.LayerBuilder`: The created model builder instance.
    """
    return get_model_repo_lazy().create_model_builder(model_name, *args, **kwargs)


def create_model_config(model_name: str, *args, **kwargs) -> "ModelConfig":
    """
    Create a model config for the given model name.

    Args:
        model_name (:obj:`str`): The name of the model for which config is to be
            created.
        *args: Additional positional arguments for config creation.
        **kwargs: Additional keyword arguments for config creation.

    Returns:
        :obj:`ModelConfig`: Model config.
    """
    return get_model_repo_lazy().create_model_config(model_name, *args, **kwargs)


def list_adapted_models(
    include_descriptions=False, filter_deprecated=False
) -> _Union[_List[str], _Dict[str, str]]:
    """
    List available models and adapted models within tamm model repository.

    Args:
        include_descriptions (:obj:`bool`, optional): Whether to include
            descriptions for each model. Defaults to `False`.
        filter_deprecated (:obj:`bool`, optional): Whether to remove deprecated
            models from the list. Defaults to `False`.

    Returns:
        :obj:`Union[List[str], Dict[str, str]]`: A list of model and adapted model
            IDs if `include_descriptions` is `False`, otherwise a dictionary with
            model and adapted model IDs as keys and their descriptions as values.
    """
    return get_model_repo_lazy().list_adapted_models(
        include_descriptions=include_descriptions, filter_deprecated=filter_deprecated
    )


def list_model_builders(
    include_descriptions=False, filter_deprecated=False
) -> _Union[_List[str], _Dict[str, str]]:
    """
    List available model configs within tamm model repository.

    Args:
        include_descriptions (:obj:`bool`, optional): Whether to include
            descriptions for each model builder. Defaults to `False`.
        filter_deprecated (:obj:`bool`, optional): Whether to remove deprecated
            model builders from the list. Defaults to `False`.

    Returns:
        :obj:`Union[List[str], Dict[str, str]]`: A list of model IDs if
            `include_descriptions` is `False`, otherwise a dictionary with model IDs
             as keys and their descriptions as values.
    """
    return get_model_repo_lazy().list_model_builders(
        include_descriptions=include_descriptions, filter_deprecated=filter_deprecated
    )


def list_model_configs(
    include_descriptions=False, filter_deprecated=False
) -> _Union[_List[str], _Dict[str, str]]:
    """
    List available model configs within tamm model repository.

    Args:
        include_descriptions (:obj:`bool`, optional): Whether to include
            descriptions for each model builder. Defaults to `False`.
        filter_deprecated (:obj:`bool`, optional): Whether to remove deprecated
            model builders from the list. Defaults to `False`.

    Returns:
        :obj:`Union[List[str], Dict[str, str]]`: A list of model IDs if
            `include_descriptions` is `False`, otherwise a dictionary with model IDs
             as keys and their descriptions as values.
    """
    return get_model_repo_lazy().list_model_configs(
        include_descriptions=include_descriptions, filter_deprecated=filter_deprecated
    )


def list_models(
    include_descriptions=False, filter_deprecated=False
) -> _Union[_List[str], _Dict[str, str]]:
    """
    List available models within tamm model repository.

    Args:
        include_descriptions (:obj:`bool`, optional): Whether to include
            descriptions for each model. Defaults to `False`.
        filter_deprecated (:obj:`bool`, optional): Whether to remove deprecated
            models from the list. Defaults to `False`.

    Returns:
        Union[List[str], Dict[str, str]]: A list of model IDs if
            `include_descriptions` is `False`, otherwise a dictionary with model IDs
            as keys and their descriptions as values.
    """
    return get_model_repo_lazy().list_models(
        include_descriptions=include_descriptions, filter_deprecated=filter_deprecated
    )


def list_available_lora_adapted_models() -> _Dict[str, _Dict[int, str]]:
    """
    Retrieve the mapping from base model to all available adapted models within
    tamm model repository.
    Returns:
        Dict[str, Dict[int, str]]: A dictionary with base model IDs as keys and
            a dictionary as values, wich has adapter rank as the keys and
            adapter model IDs as values.
    """
    return get_model_repo_lazy().list_available_lora_adapted_models()


@_tamm_warnings.deprecate(alternative="tamm.create_model_config")
def is_model_name(model_name: str) -> bool:
    """
    Check if the given name is a valid model name.

    .. admonition:: New deprecation
        :class: warning

        This API is being deprecate and will be removed in ``tamm`` 1.0.
        Please use `tamm.create_model_config(...)` and catch ``KeyError`` instead.

    Args:
        model_name (:obj:`str`): The name to be checked.

    Returns:
        :obj:`bool`: `True` if the name is a valid model name, `False` otherwise.
    """
    return get_model_repo_lazy().is_model_name(model_name)


@_tamm_warnings.deprecate(alternative="tamm.create_model_config")
def is_adapted_model_name(model_name: str) -> bool:
    """
    Check if the given name is a valid model or adapted model name.

    .. admonition:: New deprecation
        :class: warning

        This API is being deprecate and will be removed in ``tamm`` 1.0.
        Please use `tamm.create_model_config(...)` and catch ``KeyError`` instead.

    Args:
        model_name (:obj:`str`): The model name to be checked.

    Returns:
        :obj:`bool`: `True` if the name is a valid adapted model name,
            `False` otherwise.
    """
    return get_model_repo_lazy().is_adapted_model_name(model_name)


@_tamm_warnings.deprecate(alternative="tamm.create_model_config")
def is_model_builder_name(model_builder_name: str) -> bool:
    """
    Check if the given name is a valid model builder name.

    .. admonition:: New deprecation
        :class: warning

        This API is being deprecate and will be removed in ``tamm`` 1.0.
        Please use `tamm.create_model_config(...)` and catch ``KeyError`` instead.

    Args:
        model_builder_name (:obj:`str`): The builder name to be checked.

    Returns:
        :obj:`bool`: `True` if the name is a valid model builder name,
            `False` otherwise.
    """
    return get_model_repo_lazy().is_model_builder_name(model_builder_name)


@_tamm_warnings.deprecate(alternative="tamm.create_model_config")
def is_model_config_name(model_config_name: str) -> bool:
    """
    Check if the given name is a valid model config name.

    .. admonition:: New deprecation
        :class: warning

        This API is being deprecate and will be removed in ``tamm`` 1.0.
        Please use `tamm.create_model_config(...)` and catch ``KeyError`` instead.

    Args:
        model_config_name (:obj:`str`): The config name to be checked.

    Returns:
        :obj:`bool`: `True` if the name is a valid model config name,
            `False` otherwise.
    """
    return get_model_repo_lazy().is_model_config_name(model_config_name)


def get_model_repo_eager(*args, **kwargs) -> "BaseModelRepo":
    """
    Get model repo based on current environment (env variables, plugins, and calling args)

    Returns a composition of

    1. tamm-core defined model repo (currently empty, i.e., :py:class:`tamm.model_repo.model_repo.PlaceholderModelRepo`)
    2. Plugin defined model repos

    The implementation is roughly equivalent to the following pseudo-code

    .. code-block::

        repo = get_tamm_core_defined_model_repo()

        for <plugin_name> in discover_all_plugin_module_names():
            guard [Continue for loop if the following lines failed due to known exceptions]

            from <plugin_name>.model_repo import get_model_repo
            extra_repo_defined_by_plugin = get_model_repo(*args, **kwargs)
            repo += extra_repo_defined_by_plugin

        return repo

    Returns:
        Composed model repo, including extended model repos defined by plugins

    """
    # model repo defined by the core package, currently a placeholder (nothing defined)
    repo = FileModelRepo(*args, **kwargs) + URIHandlerModelRepo(*args, **kwargs)

    for plugin_name, plugin_repo in discover_and_retrieve_plugins().items():
        try:
            logger.debug("Attempt to create model repo from plugin %s", plugin_name)
            extra_repo = importlib.import_module(
                ".model_repo", package=plugin_repo
            ).get_model_repo(*args, **kwargs)
            logger.debug("Plugin %s defines model repo %s", plugin_name, extra_repo)
            logger.debug(
                "Extra model ids defined by plugin %s: %s",
                plugin_name,
                extra_repo.list_models(),
            )
            repo += extra_repo
        except (AttributeError, ModuleNotFoundError) as e:
            logger.debug(
                "Plugin %s does not extend model repo module: %s", plugin_name, e
            )
        except ImportError as e:
            logger.warning(
                "Plugin %s's model_repo cannot be imported. "
                "Models extended by plugin may not be accessible. "
                "export TAMM_LOG_LEVEL=debug to see debug info",
                plugin_name,
            )
            logger.debug("%s", e)
        except (TypeError, ValueError) as e:
            logger.warning(
                "Failed to compose model repo defined by plugin %s. "
                "Models extended by plugin may not be accessible. "
                "export TAMM_LOG_LEVEL=debug to see debug info",
                plugin_name,
            )
            logger.debug("%s", e)
        except Exception as e:
            # TODO: Now all exceptions are swallowed. Consider refactor model repo to raise error when
            #  a model expected from plugin-defined model repo fails to create. Also consider defining a
            #  new 'entry-point' (e.g., [project.entry-points.tamm.model_repo]) in package metadata to replace
            #  this trial and error discovery strategy.
            logger.warning(
                "Unhandled error occurred during discovery of model repo from plugin %s. "
                "Models extended by plugin may not be accessible. %s",
                plugin_name,
                e,
            )

    return repo



from tamm.layers.mixture_of_experts.sampler.common import (
    MixtureOfExpertsSampler,
    MixtureOfExpertsSamplerOutput,
)
from tamm.layers.mixture_of_experts.sampler.top_k import TopKMixtureOfExpertsSampler



"""
Tokenizer registry
==================

This module defines baseclass for all ``TokenizerRegistry``.

tamm plugin may extend tokenizer registry by implementing a class that inherits :py:class:`.TokenizerRegistry` and
publicize a custom tokenizer registry by exposing a factory of TokenizerRegistry  via package metadata

.. code-block:: toml

    [project.entry-points.'tamm.plugins.extras']
    tokenizer-registry = 'module.submodule:get_tokenizer_registry'

"""
from tamm.tokenizers.registry.base import ComposedTokenizerRegistry, TokenizerRegistry
from tamm.tokenizers.registry.placeholder import PlaceholderTokenizerRegistry
from tamm.tokenizers.registry.uri_handler import URIHandlerTokenizerRegistry

__all__ = [
    "TokenizerRegistry",
    "ComposedTokenizerRegistry",
    "PlaceholderTokenizerRegistry",
    "URIHandlerTokenizerRegistry",
]



"""
layers.convolution
^^^^^^^^^^^^^^^^^^

This module implements convolution-related layer building blocks.

.. autoclass:: tamm.layers.Conv2d

.. autoclass:: tamm.layers.Conv1d

.. autofunction:: tamm.layers.convolution.compute_convolution_output_dims

.. autoclass:: tamm.layers.ResNetBlock
    :members:

.. autoclass:: tamm.layers.ResNetStage
    :members:

.. autoclass:: tamm.layers.SqueezeExcitation
    :members:
"""

import math as _math
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch

from tamm.layers import activation as _activation
from tamm.layers import basic as _basic
from tamm.layers import common as _layers_common
from tamm.layers import norm as _norm
from tamm.layers import residual as _residual
from tamm.layers import sequential as _sequential
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalDeviceOrString as _OptionalDeviceOrString
from tamm.typing import OptionalDtypeOrString as _OptionalDtypeOrString
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class Conv2d(_torch.nn.Conv2d, _LayerMixin):
    """
    A |tamm| version of :class:`torch.nn.Conv2d`.  The init signature the same as the
    :mod:`torch.nn` version except ``bias`` defaults to ``False``.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _Union[int, _Tuple[int, int]],
        stride: _Union[int, _Tuple[int, int]] = 1,
        padding: _Union[str, int, _Tuple[int, int]] = 0,
        dilation: _Union[int, _Tuple[int, int]] = 1,
        groups: int = 1,
        bias: bool = False,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            device=device,
            dtype=dtype,
        )


class Conv1d(_torch.nn.Conv1d, _LayerMixin):
    """
    A |tamm| version of :class:`torch.nn.Conv1d`.  The init signature the same as the
    :mod:`torch.nn` version except ``bias`` defaults to ``False``.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _Union[_Tuple[int, int], int],
        stride: _Optional[_Union[_Tuple[int, int], int]] = None,
        padding: int = 0,
        bias: bool = False,
        groups: int = 1,
        device: _OptionalDeviceOrString = None,
        dtype: _OptionalDtypeOrString = None,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=groups,
            stride=stride,
            padding=padding,
            bias=bias,
            device=device,
            dtype=dtype,
        )


def compute_convolution_output_dims(
    *input_dims: int,
    kernel_size: _Tuple[int, ...],
    stride: _Tuple[int, ...] = 1,
    padding: _Tuple[int, ...] = 0,
) -> _torch.Size:
    """
    A helper function that computes the shape of a convolution's output given the shape
    of its input.

    Args:
        input_dims (:obj:`int`): One or more integers representing the convolution's
            input shape in the convolution dimensions (height and width in the case of
            a 2D conv).
        kernel_size (:obj:`tuple` of :obj:`int`): The ``kernel_size`` of the
            convolution.
        stride (:obj:`tuple` of :obj:`int`, optional): The ``stride`` of the
            convolution.  Defaults to ``1``.
        padding (:obj:`tuple` of :obj:`int`, optional): The ``padding`` of the
            convolution.  Defaults to ``0``.

    For ``kernel_size``, ``stride``, and ``padding``, the length should match the number
    of ``input_dims``.

    Returns:
        The shape of the convolution's output in the convolution dimensions.

    Raises:
        RuntimeError: If the kernel size exceeds the padded input size.
    """

    def normalize(inputs, *, length):
        try:
            return tuple(inputs)
        except TypeError:
            return (inputs,) * length

    ndim = len(input_dims)
    kernel_size = normalize(kernel_size, length=ndim)
    stride = normalize(stride, length=ndim)
    padding = normalize(padding, length=ndim)

    if len(kernel_size) != ndim:
        raise ValueError(f"kernel_size length ({len(kernel_size)}) != ndim ({ndim})")
    if len(stride) != ndim:
        raise ValueError(f"stride length ({len(stride)}) != ndim ({ndim})")
    if len(padding) != ndim:
        raise ValueError(f"padding length ({len(padding)}) != ndim ({ndim})")

    shapes = zip(input_dims, kernel_size, stride, padding)
    result = []
    for dim_i, kernel_size_i, stride_i, padding_i in shapes:
        numerator = dim_i + 2 * padding_i - kernel_size_i + 1
        if numerator <= 0:
            raise RuntimeError("Kernel size can't exceed the input size")
        output_dim = _math.ceil(numerator / stride_i)
        result.append(output_dim)
    return _torch.Size(result)


class ResNetBlock(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    A generic ResNet block.  This supports both basic block and bottleneck layers as
    well as ResNet variants such as RegNet.  The layer expects NCHW image tensor inputs.
    """

    def __init__(
        self,
        *,
        convolution_0: _ModuleOrBuilder,
        norm_0: _ModuleOrBuilder,
        activation_0: _activation.ActivationSpecType,
        convolution_1: _ModuleOrBuilder,
        norm_1: _ModuleOrBuilder,
        activation_1: _Optional[_activation.ActivationSpecType] = None,
        squeeze_excitation_1: _OptionalModuleOrBuilder = None,
        convolution_2: _OptionalModuleOrBuilder = None,
        norm_2: _OptionalModuleOrBuilder = None,
        squeeze_excitation_2: _OptionalModuleOrBuilder = None,
        residual_connection: _OptionalModuleOrBuilder = None,
    ):
        layers = {
            "convolution_0": convolution_0,
            "norm_0": norm_0,
            "activation_0": _activation.create_activation_layer(activation_0),
            "convolution_1": convolution_1,
            "norm_1": norm_1,
            "activation_1": _activation.create_activation_layer(activation_1),
            "squeeze_excitation_1": squeeze_excitation_1,
            "convolution_2": convolution_2,
            "norm_2": norm_2,
            "squeeze_excitation_2": squeeze_excitation_2,
        }
        super().__init__(layers, residual_connection=residual_connection)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        input_dim: int,
        hidden_dim: _Optional[int] = None,
        output_dim: _Optional[int] = None,
        kernel_size: _Union[int, _Tuple[int, int]] = 3,
        stride: _Union[int, _Tuple[int, int]] = 1,
        padding: _Union[int, _Tuple[int, int]] = 1,
        groups: int = 1,
        bias: bool = False,
        activation: str = "relu",
        norm: str = "rms_norm",
        norm_eps: float = 1e-5,
        norm_bias: bool = False,
        apply_squeeze_excitation: bool = False,
        squeeze_excitation_reduced_dim: _Optional[int] = None,
        squeeze_excitation_bias: bool = False,
        squeeze_excitation_activation: str = "glu",
    ) -> _layers_common.LayerBuilder:
        """
        Creates a builder for a `ResNet <https://arxiv.org/abs/1512.03385>`__ bottleneck
        layer.

        Args:
            input_dim (:obj:`int`): The number of input channels.
            hidden_dim (:obj:`int`, optional): The number of hidden channels.  If
                ``None``, this defaults to ``input_dim``.
            output_dim (:obj:`int`, optional): The number of output channels.  If
                ``None``, this defaults to ``input_dim``.
            kernel_size (:obj:`int` or :obj:`tuple`, optional): The ``kernel_size``
                argument for ``convolution_1``.  Defaults to ``3``.
            stride (:obj:`int`): The ``stride`` argument for ``convolution_1``.
                Defaults to ``1``.
            padding (:obj:`int` or :obj:`tuple`): The ``padding`` argument for
                ``convolution_1``.  Defaults to ``1``.
            groups (:obj:`int`): The number of groups for ``convolution_1``.  Defaults
                to ``1``.
            bias (:obj:`bool`, optional): A flag for including ``bias`` parameters in
                convolution layers.  Defaults to ``False``.
            activation (:obj:`str`, optional): The activation type.  Defaults to
                ``"relu"``.
            norm (:obj:`str`, optional): A norm "spec" argument for
                :func:.`create_norm_builder`.  Defaults to ``"rms_norm"``.
            norm_eps (:obj:`float`, optional): The epsilon value for norm layers.
                Defaults to ``1e-5``.
            norm_bias (:obj:`bool`, optional): Whether or not to use a ``bias``
                parameter in norm layers.  Defaults to ``False``.
            apply_squeeze_excitation (:obj:`bool`, optional): A flag for including
                a :class:`SqueezeExcitation` layer after ``norm_2`` when
                ``True``.  Defaults to ``False``.
            squeeze_excitation_reduced_dim (:obj:`int`, optional): The reduced dimension
                for squeeze excitation.  If ``None``, this defaults to
                ``ceil(hidden_dim / 16)``.
            squeeze_excitation_bias (:obj:`bool`, optional): The bias option for the
                squeeze-excitation layer.  Defaults to ``False``.
            squeeze_excitation_activation (:obj:`str`, optional): The gated activation
                type for the squeeze-excitation layer.  Defaults to ``"glu"``.

        Returns:
            The newly created :obj:`LayerBuilder`.
        """

        # pylint: disable=too-many-locals

        if hidden_dim is None:
            hidden_dim = input_dim
        if output_dim is None:
            output_dim = input_dim

        builder_kwargs = {}

        for idx in range(3):
            out_channels = output_dim if idx == 2 else hidden_dim

            conv = Conv2d.Builder(
                in_channels=input_dim if idx == 0 else hidden_dim,
                out_channels=out_channels,
                kernel_size=1,
                bias=bias,
            )
            if idx == 1:
                conv.kernel_size = kernel_size
                conv.stride = stride
                conv.padding = padding
                conv.groups = groups
            builder_kwargs[f"convolution_{idx}"] = conv

            norm_builder = _norm.create_norm_builder(
                (out_channels,),
                norm,
                dim=1,
                eps=norm_eps,
                bias=norm_bias,
            )
            builder_kwargs[f"norm_{idx}"] = norm_builder

            if idx != 2:
                builder_kwargs[f"activation_{idx}"] = activation

        if apply_squeeze_excitation:
            builder_kwargs["squeeze_excitation_2"] = SqueezeExcitation.create_builder(
                input_dim=hidden_dim,
                reduced_dim=squeeze_excitation_reduced_dim,
                bias=squeeze_excitation_bias,
                activation=activation,
                gated_activation=squeeze_excitation_activation,
            )

        if stride != 1 or input_dim != output_dim:
            # downsample
            shortcut_layers = {
                "convolution": Conv2d.Builder(
                    in_channels=input_dim,
                    out_channels=output_dim,
                    kernel_size=1,
                    stride=stride,
                    bias=bias,
                ),
                "norm": _norm.create_norm_builder(
                    (out_channels,),
                    norm,
                    dim=1,
                    eps=norm_eps,
                    bias=norm_bias,
                ),
            }
            shortcut = _sequential.Sequential.Builder(shortcut_layers)
        else:
            shortcut = None
        residual_connection = _residual.ShortcutAddActResidualConnection.Builder(
            shortcut=shortcut,
            activation=activation,
        )

        return cls.Builder(**builder_kwargs, residual_connection=residual_connection)


class ResNetStage(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    A sequence of :obj:`ResNetBlock` layers.

    Args:
        blocks (:obj:`list`): A list of builders for ResNet blocks.
    """

    # pylint: disable=too-many-locals

    def __init__(self, blocks: _List[_ModuleOrBuilder]):
        named_layers = {f"block_{idx}": block for idx, block in enumerate(blocks)}
        super().__init__(named_layers)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        input_dim: int,
        hidden_dim: _Optional[int] = None,
        output_dim: _Optional[int] = None,
        num_blocks: int = 1,
        kernel_size: _Union[int, _Tuple[int, int]] = 3,
        stride: _Union[int, _Tuple[int, int]] = 1,
        padding: _Union[int, _Tuple[int, int]] = 1,
        dim_per_group: _Optional[int] = None,
        activation: str = "relu",
        norm: str = "rms_norm",
        norm_eps: float = 1e-5,
        norm_bias: bool = False,
        bias: bool = False,
        apply_squeeze_excitation: bool = False,
        squeeze_excitation_reduced_dim: _Optional[int] = None,
        squeeze_excitation_bias: bool = False,
        squeeze_excitation_activation: str = "glu",
    ) -> _layers_common.LayerBuilder:
        """
        Creates a :obj:`.LayerBuilder` for a :obj:`.ResNetStage` with bottleneck blocks.

        Args:
            num_blocks (:obj:`int`, optional): The number of blocks in the stage.
                Defaults to ``1``.
            dim_per_group (:obj:`int`, optional): The ``input_dim`` of each block
                divided by the number of convolution groups in the block's
                ``convolution_1``.  Defaults to ``None``, in which case each block uses
                only one convolution group.

        Please see :meth:`.ResNetBlock.create_builder` for a description of the
        remaining arguments.

        Returns:
            The newly created :obj:`LayerBuilder`.
        """
        blocks = []
        for _ in range(num_blocks):
            if dim_per_group is None:
                groups = 1
            else:
                groups = input_dim // dim_per_group

            block = ResNetBlock.create_builder(
                input_dim=input_dim,
                hidden_dim=hidden_dim,
                output_dim=output_dim,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=groups,
                activation=activation,
                norm=norm,
                norm_eps=norm_eps,
                norm_bias=norm_bias,
                bias=bias,
                apply_squeeze_excitation=apply_squeeze_excitation,
                squeeze_excitation_reduced_dim=squeeze_excitation_reduced_dim,
                squeeze_excitation_bias=squeeze_excitation_bias,
                squeeze_excitation_activation=squeeze_excitation_activation,
            )
            blocks.append(block)
            input_dim = output_dim
            stride = 1
        return cls.Builder(blocks)

    @classmethod
    def create_regnet_builder(cls, **kwargs) -> _layers_common.LayerBuilder:
        """
        Creates a builder for a `RegNet <https://arxiv.org/abs/2003.13678>`__ stage
        with bottleneck layers.  Following ``timm``, this is the same as a ResNet stage
        layer except we set ``apply_squeeze_excitation=True`` by default and the
        :class:`.SqueezeExcitation` layers lies between ``activation_1`` and
        ``convolution_2`` within each block.  The defaults for ``hidden_dim`` and
        ``squeeze_excitation_reduced_dim`` may also differ from ``timm``, so it is
        recommended to pass these values explicitly.

        Args:
            **kwargs: Keyword arguments to :meth:`.ResNetStage.create_builder`.
        """

        builder = cls.create_builder(apply_squeeze_excitation=True, **kwargs)
        for block_builder in builder.blocks:
            block_builder.squeeze_excitation_1 = block_builder.squeeze_excitation_2
            block_builder.squeeze_excitation_2 = None
        return builder


class SqueezeExcitation(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    A `Squeeze-and-Excitation <https://arxiv.org/abs/1709.01507>`__ block.  This is
    a gated activation layer with gate inputs resulting from a ``squeeze``,
    ``reduce_transform``, ``activation``, and ``expand_transform`` layer sequence.
    This layer expects NCHW image tensor inputs.
    """

    def __init__(
        self,
        *,
        gated_activation: _activation.ActivationSpecType,
        squeeze: _ModuleOrBuilder,
        reduce_transform: _ModuleOrBuilder,
        activation: _activation.ActivationSpecType,
        expand_transform: _ModuleOrBuilder,
    ):
        layers = {
            "squeeze": squeeze,
            "reduce_transform": reduce_transform,
            "activation": _activation.create_activation_layer(activation),
            "expand_transform": expand_transform,
        }
        residual_connection = _residual.GatedActivationResidualConnection(
            gated_activation
        )
        super().__init__(layers, residual_connection=residual_connection)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        input_dim: int,
        reduced_dim: _Optional[int] = None,
        bias: bool = False,
        activation: str = "relu",
        gated_activation: str = "glu",
    ) -> _layers_common.LayerBuilder:
        """
        Creates a builder for squeeze-and-excitation layers.

        Args:
            input_dim (:obj:`int`): The number of input channels.
            reduced_dim (:obj:`int`, optional): The reduced dimension for excitation.
                If ``None``, this defaults to ``ceil(input_dim / 16)``.
            bias (:obj:`bool`, optional): A flag for including ``bias`` parameters in
                convolution layers.  Defaults to ``False``.
            activation (:obj:`str`, optional): The activation type.  Defaults to
                ``"relu"`` for excitation.
            gated_activation (:obj:`str`, optional): The gated activation type.
                Defaults to ``"glu"``.

        Returns:
            The newly created :obj:`LayerBuilder`.
        """

        if reduced_dim is None:
            reduced_dim = _math.ceil(input_dim / 16)
        conv_kwargs = {"kernel_size": 1, "bias": bias}

        reduce_transform = Conv2d.Builder(
            in_channels=input_dim, out_channels=reduced_dim, **conv_kwargs
        )
        expand_transform = Conv2d.Builder(
            in_channels=reduced_dim, out_channels=input_dim, **conv_kwargs
        )

        return cls.Builder(
            squeeze=_basic.Mean.Builder(dim=(2, 3), keepdim=True),
            reduce_transform=reduce_transform,
            activation=activation,
            expand_transform=expand_transform,
            gated_activation=gated_activation,
        )



import abc as _abc

import torch as _torch

from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class MixtureOfExpertsSamplerOutput:
    """
    Result dataclass for :obj:`MixtureOfExpertsSampler` layers.

    Args:
        expert_assignments (:obj:`torch.Tensor`): An integer tensor that
            maps tokens to chosen experts.  The expected shape is
            ``(*batch_shape, seq_len, num_experts_per_token)``.
        assignment_weights (:obj:`torch.Tensor`): A floating point tensor with
            the same shape as ``expert_assignments``.  These are weights
            for combining expert outputs for each token.
    """

    expert_assignments: _torch.Tensor
    assignment_weights: _torch.Tensor


class MixtureOfExpertsSampler(_torch.nn.Module, _abc.ABC):
    """An abstract base class for MoE samplers."""

    @_abc.abstractmethod
    def forward(self, probabilities: _torch.Tensor) -> MixtureOfExpertsSamplerOutput:
        """
        Transforms routing probabilities into weighted expert assignments.

        Args:
            probabilities (:obj:`torch.Tensor`): The output of the router's
                softmax op.  The expected shape is
                ``(*batch_shape, seq_len, num_experts)``.

        Returns a :obj:`MixtureOfExpertsSamplerOutput`.
        """



"""
token_generators.sampling
^^^^^^^^^^^^^^^^^^^^^^^^^
"""

from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn

from tamm.generation.token_generators._vanilla import _VanillaTokenGenerator
from tamm.generation.token_generators.base import (
    TokenGeneratorOutput as _TokenGeneratorOutput,
)
from tamm.generation.token_generators.next_token_logic import (
    NextTokenLogic as _NextTokenLogic,
)
from tamm.generation.token_generators.next_token_logic import (
    NextTokenLogicOutput as _NextTokenLogicOutput,
)
from tamm.generation.token_generators.utils import KVCacheType as _KVCacheType
from tamm.generation.token_generators.utils import process_logit as _process_logit
from tamm.utils.torch_utils import (
    torch_exportable_dataclass as _torch_exportable_dataclass,
)


class SamplingNextTokenLogic(_NextTokenLogic):
    """
    Generates next token by transforming the logits into a probability distribution
    using the sampling parameters (temperature scaling, top_k and top_p filtering)
    and drawing a random sample from this distribution.

    Args:
        temperature (:obj:`float`): A scaling factor applied to the logits to adjust the randomness
          of the predictions. Defaults to ``None``.
        top_k (:obj:`int`): The number of highest probability tokens to consider when filtering the logits.
          if ``top_k`` is greater than 0, only the top k tokens with highest probability are retained.
          Defaults to ``None``.
        top_p (:obj:`float`): Cumulative probability threshold for nucleus sampling. Defaults to ``None``.
    """

    def __init__(
        self,
        temperature: _Optional[float] = None,
        top_k: _Optional[int] = None,
        top_p: _Optional[float] = None,
    ):
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p

    def compute_probabilities(self, logits: _torch.Tensor) -> _torch.Tensor:
        """
        Returns a probability distribution by transforming logits using the sampling
        parameters.
        """
        _, probs = _process_logit(
            logits, temperature=self.temperature, top_k=self.top_k, top_p=self.top_p
        )
        return probs

    def __call__(
        self,
        logits: _torch.Tensor,
        return_probabilities: bool = False,
        rng: _Optional[_torch.Generator] = None,
    ) -> _NextTokenLogicOutput:
        probs = self.compute_probabilities(logits)
        next_tokens = _torch.multinomial(probs, num_samples=1, generator=rng)
        if return_probabilities:
            return _NextTokenLogicOutput(token_ids=next_tokens, probabilities=probs)
        return _NextTokenLogicOutput(token_ids=next_tokens)


@_torch_exportable_dataclass
class SamplingTokenGeneratorOutput(_TokenGeneratorOutput):
    """
    Output generated from :py:class:`SamplingTokenGenerator`, consisting of token_ids and
    corresponding logits, if that option is enabled.

    Args:
        token_ids (:py:class:`_torch.Tensor`): Tokens produced by the sampling generator.
          Does not include ``input_ids``.
        logits (optional :py:class:`_torch.Tensor`): Logits corresponding to the generated tokens
        probabilities (optional :py:class:`_torch.Tensor`): Probability distributions corresponsing to
         the token ids.
    """

    token_ids: _torch.Tensor
    logits: _Optional[_torch.Tensor] = None
    probabilities: _Optional[_torch.Tensor] = None


class SamplingTokenGenerator(_VanillaTokenGenerator):
    """
    Token generator that uses sampling decoding strategy.

    Args:
        model (:obj:`nn.Module`): The model instance.
        eos_id (:obj:`Union[int, torch.Tensor, List[torch.Tensor]]`): End of sequence token ID(s).
        pad_id (:obj:`Union[int, torch.Tensor]`): Padding token ID.
        temperature (:obj:`float`): A scaling factor applied to the logits to adjust the randomness
          of the predictions. Defaults to ``None``.
        top_k (:obj:`int`): The number of highest probability tokens to consider when filtering the logits.
          if ``top_k`` is greater than 0, only the top k tokens with highest probability are retained.
          Defaults to ``None``.
        top_p (:obj:`float`): Cumulative probability threshold for nucleus sampling. Defaults to ``None``.
        cache_type (:obj:`KVCacheType`, optional): Type of KV cache to use. Defaults to KVCacheType.V0.
        cache_dtype (:obj:`torch.dtype`, optional): Data type for KV cache. Defaults to None.
        output_logits (:obj:`bool`): When set to ``True``, :py:meth:`generate` returns logits
          corresponding to the generated ``token_ids``.
        output_probabilities (:obj:`bool`): When set to ``True``, :py:meth:`generate` returns probability
          distributions corresponding to the generated ``token_ids``.
    """

    def __init__(
        self,
        model: _nn.Module,
        eos_id: _Union[int, _torch.Tensor, _List[_torch.Tensor]],
        pad_id: _Union[int, _torch.Tensor],
        temperature: _Optional[float] = None,
        top_k: _Optional[int] = None,
        top_p: _Optional[float] = None,
        cache_type: _Optional[_KVCacheType] = _KVCacheType.V0,
        cache_dtype: _Optional[_torch.dtype] = None,
        output_logits: bool = False,
        output_probabilities: bool = False,
    ):
        next_token_logic = SamplingNextTokenLogic(
            temperature=temperature, top_k=top_k, top_p=top_p
        )
        # pylint: disable=duplicate-code
        super().__init__(
            model=model,
            next_token_logic=next_token_logic,
            eos_id=eos_id,
            pad_id=pad_id,
            cache_type=cache_type,
            cache_dtype=cache_dtype,
            output_logits=output_logits,
            output_probabilities=output_probabilities,
        )

    @_torch.inference_mode()
    def generate(
        self,
        input_ids: _torch.Tensor,
        max_new_tokens: int,
        seed: _Optional[int] = None,
    ) -> SamplingTokenGeneratorOutput:
        token_ids, return_logits, return_probs = self._generate(
            input_ids, max_new_tokens, seed
        )
        return SamplingTokenGeneratorOutput(
            token_ids=token_ids, logits=return_logits, probabilities=return_probs
        )



"""
models.afm_text
===============

.. autoclass:: tamm.models.afm_text.AFMTextV7
    :show-inheritance:
    :members:
"""

import copy as _copy
from typing import Optional as _Optional
from typing import Union as _Union

from tamm import layers as _layers
from tamm.models import common as _models_common


class AFMTextV7(_layers.CausalLMTransformer, _models_common.ModelMixin):
    """
    AFM Text v7 model.  This is a decoder-only causal LM with:

    * Key/value reuse across layers
    * Rotary positional embeddings
    * Grouped query attention
    * Tied embedding and head layers
    """

    @classmethod
    def create_basic_builder(
        cls,
        *,
        vocab_size: int = 153600,
        hidden_dim: int = 2048,
        num_layers: int = 56,
        num_kv_reuse_layers: int = 21,
        num_heads: int = 16,
        num_kv_heads: _Optional[int] = 2,
        hidden_dim_scale_factor: float = 3.25,
        rope_theta: float = 500000.0,
        attention_dropout_p: float = 0.0,
        hidden_dropout_p: float = 0.0,
        include_loss_layer: bool = False,
        output_hidden_states: bool = False,
        output_attentions: bool = False,
        sdpa_implementation: _Union[
            _layers.attention.SDPAImplementationChoice, str
        ] = "auto",
    ):
        """
        Creates and returns a configured builder for the model.

        Args:
            vocab_size (:obj:`int`): The vocab size.
            hidden_dim (:obj:`int`): The number of features per token.
            num_layers (:obj:`int`): The number of transformer layers.
            num_kv_reuse_layers (:obj:`int`):  The number of KV reuse attention layers.
                If larger than ``0``, the first ``num_layers - num_kv_reuse_layers``
                layers are regular layers, while the final ``num_kv_reuse_layers`` layers
                reuse ``keys`` and ``values`` states from the last regular layer.
                Defaults to ``21``.
            num_heads (:obj:`int`): The number of attention heads.
            num_kv_heads (:obj:`int`, optional): The number of key-value attention
                heads for grouped query attention.  If ``None``, this defaults to
                ``num_heads``.  If not ``None``, this value must evenly divide
                ``num_heads``.
            hidden_dim_scale_factor (:obj:`float`, optional): For feed-forward layers,
                the ratio of the layer's hidden dimension to the layer's input
                dimension.  Defaults to ``3.25``.
            rope_theta (:obj:`float`, optional): The theta value for rotary positional
                embeddings.  Defaults to ``500000``.
            attention_dropout_p (:obj:`float`, optional): The dropout rate for
                attention probabilities.  Defaults to 0.
            hidden_dropout_p (:obj:`float`, optional): The dropout rate for hidden
                states.
            include_loss_layer (:obj:`bool`): Flag for including a cross-entropy loss
                layer in the model when ``True``.  Defaults to ``False``.  Unlike
                Hugging Face, this layer does not shift the labels.
            output_hidden_states (:obj:`bool`): Flag that when ``True`` results in the
                inclusion of hidden states in the final output of the model.
                Specifically, this is a list of ``num_layers + 1`` tensors that includes
                the initial embeddings and the output of each layer.  Defaults to
                ``False``.
            output_attentions (:obj:`bool`): Flag that when ``True`` results in the
                inclusion of attentions in the output of the model.
                Specifically, this is a list of ``num_layers`` tensors that includes
                attentions of each layer.  Defaults to ``False``.
            sdpa_implementation (:obj:`SDPAImplementationChoice` member or :obj:`str`):
                The attention implementation to use.  The layer uses ``AUTO`` by
                default, which attempts to select the most efficient implementation
                available.  If a :obj:`str`, this argument must equal the name of a
                :class:`.SDPAImplementationChoice` member (case-insensitive).

        Returns:
            The configured :obj:`.AFMTextModelBuilder`.
        """

        # pylint: disable=too-many-locals

        builder = cls.Builder()

        builder.embedding = _layers.Embedding.Builder(
            vocab_size, hidden_dim, padding_idx=0
        )
        builder.attention_mask = _layers.transformer.AttentionMask.Builder(
            is_causal=True
        )

        dim_per_head = hidden_dim // num_heads
        builder.positional_encoding = _layers.RotaryPositionalEmbedding.Builder(
            dim_per_head=dim_per_head, theta=rope_theta
        )

        norm = _layers.RMSNorm.Builder([hidden_dim])

        attention = _layers.attention.TransformerAttention.create_builder(
            target_dim=hidden_dim,
            num_heads=num_heads,
            num_kv_heads=num_kv_heads,
            norm=norm,
            apply_rope=True,
            apply_qk_norm=True,
            attention_dropout_p=attention_dropout_p,
            hidden_dropout_p=hidden_dropout_p,
            output_attentions=output_attentions,
            sdpa_implementation=sdpa_implementation,
        )
        attention.qk_norm.query_norm = _layers.RMSNorm.Builder([dim_per_head])
        attention.qk_norm.key_norm = _layers.RMSNorm.Builder([dim_per_head])

        feed_forward = _layers.feed_forward.TransformerFeedForward.create_builder(
            input_dim=hidden_dim,
            hidden_dim=round(hidden_dim * hidden_dim_scale_factor),
            norm=norm,
            activation=_layers.activation.SwiGLU.Builder(),
            activation_dropout_p=hidden_dropout_p,
            output_dropout_p=hidden_dropout_p,
        )

        layer = _layers.transformer.TransformerLayer.Builder(
            attention=attention, feed_forward=feed_forward
        )

        if num_kv_reuse_layers == 0:
            builder.layers = _layers.UniformTransformerLayerSequence.Builder(
                layer, num_layers=num_layers, output_hidden_states=output_hidden_states
            )
        else:
            builder.layers = cls._create_layers_with_kv_reuse(
                layer=layer,
                num_kv_reuse_layers=num_kv_reuse_layers,
                num_layers=num_layers,
                hidden_dim=hidden_dim,
                output_hidden_states=output_hidden_states,
            )

        builder.output_norm = norm
        builder.output_transform_build_mode = "pass_embedding"
        builder.output_transform = _layers.TiedWeightLinear.Builder(
            parameter_name="weight"
        )

        if include_loss_layer:
            builder.loss = _layers.FlattenedCrossEntropyLoss.Builder()

        return builder

    @staticmethod
    def _create_layers_with_kv_reuse(
        layer, *, num_kv_reuse_layers, num_layers, hidden_dim, output_hidden_states
    ):
        num_regular_layers = num_layers - num_kv_reuse_layers
        if num_regular_layers <= 0:
            raise ValueError(
                "num_regular_layers must be at least 1, but it has value "
                f"{num_regular_layers}"
            )

        segment_0 = [_copy.deepcopy(layer) for _ in range(num_regular_layers)]
        segment_0[-1].attention.scaled_dot_product_attention.output_kv_state = True
        segment_0 = _layers.TransformerLayerSequence.Builder(
            *segment_0, output_hidden_states=output_hidden_states
        )

        kv_reuse_layer = _copy.deepcopy(layer)
        q_transform = _layers.Linear.Builder(hidden_dim, hidden_dim, bias=False)
        sdpa = kv_reuse_layer.attention.scaled_dot_product_attention
        kv_reuse_layer.attention = _layers.KVReuseTransformerAttention.Builder(
            norm=kv_reuse_layer.attention.norm,
            q_transform=q_transform,
            rope_transform=kv_reuse_layer.attention.rope_transform,
            q_norm=kv_reuse_layer.attention.qk_norm,
            scaled_dot_product_attention=sdpa,
            output_transform=kv_reuse_layer.attention.output_transform,
            dropout=kv_reuse_layer.attention.dropout,
            residual_connection=kv_reuse_layer.attention.residual_connection,
        )
        if kv_reuse_layer.attention.q_norm is not None:
            kv_reuse_layer.attention.q_norm.key_norm = None
        segment_1 = _layers.UniformTransformerLayerSequence.Builder(
            kv_reuse_layer,
            num_layers=num_kv_reuse_layers,
            output_hidden_states=output_hidden_states,
        )

        return _layers.KVReuseTransformerLayerSequence.Builder(
            segment_0=segment_0, segment_1=segment_1
        )



import dataclasses

import torch.nn as _nn

from tamm.runtime_configuration import rc as _rc
from tamm.utils.optional_bool import OptionalBool


def init_marker(module: _nn.Module, overwrite=False):
    """
    Attaches default module marker to a ``nn.Module``. No-op if ``module`` already
    has an attribute named :py:obj:`tamm.rc.constants.MARKER_ATTRIBUTE`, unless
    ``overwrite=True``

    Args:
        module (:obj:`nn.Module`): Any ``nn.Module``
        overwrite: Force overwrite if ``True``

    """
    if overwrite or not hasattr(module, _rc.MARKER_ATTRIBUTE):
        setattr(module, _rc.MARKER_ATTRIBUTE, ModuleMarker())


def update_marker(module: _nn.Module, **attributes):
    """
    'Upserts' a dictionary of ``attributes`` to a ``nn.Module``.
    If ``module`` does not have an attribute named
    :py:obj:`tamm.rc.constants.MARKER_ATTRIBUTE`, the attribute will be created
    before updating

    Args:
        module (:obj:`nn.Module`): Any ``nn.Module``
        **attributes: Keyword markers to upsert.
            Must be :class:`ModuleMarker` attributes

    Returns: None
    """
    init_marker(module, overwrite=False)
    new_attributes = dataclasses.replace(
        getattr(module, _rc.MARKER_ATTRIBUTE), **attributes
    )
    setattr(module, _rc.MARKER_ATTRIBUTE, new_attributes)


def get_marker(module: _nn.Module) -> "ModuleMarker":
    """
    Reads extra markers from a ``nn.Module``.

    Raises ``AttributeError`` if ``module`` does not have extra markers

    Args:
        module (:obj:`nn.Module`): Any ``nn.Module``

    Returns: :class:`ModuleMarker`
    """
    init_marker(module, overwrite=False)
    return getattr(module, _rc.MARKER_ATTRIBUTE)


@dataclasses.dataclass
class ModuleMarker:
    """
    A dataclass which can be optionally attached to a ``nn.Module`` as an attribute
    named :py:obj:`tamm.rc.constants.MARKER_ATTRIBUTE`

    .. tip::

        Use :meth:`update_marker`, :meth:`init_marker` and :meth:`get_marker` to
        interact with the attribute.

    """

    freeze_params_hook_applied: "OptionalBool" = OptionalBool.NOTSET
    """
    Indicates whether builder has applied the 'freeze parameters' post hook to the
    module
    """

    weights_initialized: "bool" = False
    """
    Indicates whether a submodule has already initialized its weights.
    """

    def __post_init__(self):
        self.freeze_params_hook_applied = OptionalBool(self.freeze_params_hook_applied)



import abc as _abc
import enum as _enum
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm import _helpers
from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class MixtureOfExpertsDispatchOutput:
    """
    A dataclass for holding the outputs of a :obj:`MixtureOfExpertsDispatcher` in
    dispatch mode.

    Args:
        experts_arg:  The first positional argument to the experts layer,
            typically a tensor containing dispatched tokens.  The type and shape
            of this object depends on the dispatcher type.
        experts_kwargs (:obj:`dict`): Keyword arguments to the experts layer.
            The contents may depend on the dispatcher type.
        combine_input: An auxiliary input for calling the dispatcher in combine
            mode (after computing the outputs of experts).
    """

    experts_arg: _Any

    experts_kwargs: _Dict[str, _Any]

    combine_input: _Any


class MixtureOfExpertsDispatcherForwardMode(str, _enum.Enum):
    """
    An :obj:`Enum` for specifying the forward mode of a
    :obj:`MixtureOfExpertsDispatcher`.
    """

    DISPATCH = "DISPATCH"
    """Dispatch tokens to the experts."""

    COMBINE = "COMBINE"
    """Combine outputs from the experts."""


class MixtureOfExpertsDispatcher(_torch.nn.Module, _abc.ABC):
    """
    Base class for MoE layers that dispatch inputs to experts and later combine outputs
    from the experts.

    Example:

        .. code-block:: python

            # Dispatch tokens:
            dispatch_output = dispatcher(
                inputs, mode="dispatch", expert_assignments=expert_assignments
            )

            # Compute expert outputs (interface depends on the dispatcher type):
            experts_output = experts(
                dispatch_output.experts_arg, **dispatch_output.experts_kwargs
            )

            # Combine expert outputs:
            hidden_states = dispatcher(
                experts_output,
                mode="combine",
                assignment_weights=assignment_weights,
                combine_input=dispatch_output.combine_input,
            )
    """

    def __init__(self):
        super().__init__()

    @_abc.abstractmethod
    def _dispatch(
        self, hidden_states: _torch.Tensor, *, expert_assignments: _torch.Tensor
    ) -> MixtureOfExpertsDispatchOutput:
        """
        Dispatches tokens to experts, returning a
        :obj:`MixtureOfExpertsDispatchOutput`.

        Args:
            hidden_states (:obj:`torch.Tensor`): Input hidden states with shape
                ``(*batch_dim, seq_len, hidden_dim)``.
            expert_assignments (:obj:`torch.Tensor`): An integer tensor with shape
                ``(*batch_dim, seq_len, num_experts_per_token)`` that contains the
                indices of the experts to evaluate for each token.
        """

    @_abc.abstractmethod
    def _combine(
        self,
        experts_output: _Any,
        *,
        assignment_weights: _torch.Tensor,
        combine_input: _Any,
    ) -> _torch.Tensor:
        """
        Combines the outputs of experts, returning a :obj:`torch.Tensor`.

        Args:
            experts_output: The output of the experts layer.  This is typically
                a tensor, but the type or shape may vary depending on the
                dispatcher subclass.
            assignment_weights (:obj:`torch.Tensor`): A floating point tensor with
                the same shape as ``expert_assignments`` from dispatch mode.
                These are weights for combining expert outputs.
            combine_input (:obj:`torch.Tensor`): Auxiliary combine inputs
                produced by the dispatch call.
        """

    def forward(
        self,
        inputs: _Any,
        *,
        mode: MixtureOfExpertsDispatcherForwardMode,
        expert_assignments: _Optional[_torch.Tensor] = None,
        assignment_weights: _Optional[_torch.Tensor] = None,
        combine_input: _Any = None,
    ):
        """
        Either dispatches tokens to experts (if in dispatch mode) or combines
        outputs from experts (if in combine mode).

        Args:
            inputs: In dispatch mode, these are the tensor inputs (hidden states)
                to the MoE layer.  In combine mode, these are the outputs of the
                experts layer.
            mode (:obj:`str` or :obj:`MixtureOfExpertsDispatcherForwardMode`): The
                mode (either ``"dispatch"`` or ``"combine"``).
            expert_assignments (:obj:`torch.Tensor`): For dispatch mode only. An
                integer tensor that maps tokens to chosen experts.  The expected
                shape is ``(*batch_shape, seq_len, num_experts_per_token)``.
            assignment_weights (:obj:`torch.Tensor`): For combine mode only.  A
                floating point tensor with the same shape as
                ``expert_assignments``.  These are weights for combining expert
                outputs.
            combine_input (:obj:`torch.Tensor`): For combine mode only.
                Auxiliary combine inputs produced by the dispatch call.
        """

        mode = _helpers.get_enum_member_from_name(
            MixtureOfExpertsDispatcherForwardMode, mode
        )

        if mode is MixtureOfExpertsDispatcherForwardMode.DISPATCH:
            return self._dispatch(inputs, expert_assignments=expert_assignments)

        if mode is MixtureOfExpertsDispatcherForwardMode.COMBINE:
            return self._combine(
                inputs,
                assignment_weights=assignment_weights,
                combine_input=combine_input,
            )

        raise RuntimeError(f"Mode {mode} not recognized")



import sys as _sys
from typing import Callable as _Callable
from typing import List as _List
from typing import Mapping as _Mapping

from tamm._cli._listers.converters import ConverterLister
from tamm._cli._listers.models import AdaptedModelsLister, ModelsLister
from tamm._cli._listers.preprocessors import PreprocessorLister
from tamm._cli._listers.tokenizers import TokenizerLister

__all__ = [
    "AdaptedModelsLister",
    "ConverterLister",
    "ModelsLister",
    "TokenizerLister",
    "PreprocessorLister",
]


def get_registered_listers() -> _Mapping[str, _Callable]:
    """
    Automatically discover all listers from the imported names of this module

    Returns:
        A dictionary maps lister name to :class:`tamm._cli._listers._base.BaseLister`
    """
    listers = [getattr(_sys.modules[__name__], lister) for lister in __all__]
    return {lister.name: lister for lister in listers}


def get_registered_listers_names() -> _List[str]:
    return sorted(list(get_registered_listers().keys()))



"""
transformer.positional_encoding
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This submodule implements positional encoding layers for :class:`.TransformerStack`.
These layers encode positions into embeddings and attention masks prior to the sequence
of transformer layers.
Unlike :mod:`tamm.layers.positional_encoding`, this submodule is specific to
:class:`.TransformerStack`.

Base class
----------

.. autoclass:: tamm.layers.TransformerPositionalEncoding
    :show-inheritance:
    :members: forward

.. autoclass:: TransformerPositionalEncodingOutput


Positional encodings
--------------------

.. autoclass:: tamm.layers.ALiBiPositionalEncoding
    :show-inheritance:

.. autoclass:: tamm.layers.NoPositionalEncoding
    :show-inheritance:

.. autoclass:: tamm.layers.RotaryPositionalEmbedding
    :show-inheritance:

.. autoclass:: tamm.layers.SinusoidalPositionalEmbedding
    :show-inheritance:

.. autoclass:: tamm.layers.SlidingWindowPositionalEncoding
    :show-inheritance:

.. autoclass:: tamm.layers.TransformerAbsolutePositionalEmbedding
    :show-inheritance:
    :members: create_builder


Composite positional encodings
------------------------------

.. autoclass:: tamm.layers.SequentialPositionalEncoding
    :show-inheritance:


Secondary positional encodings
------------------------------

.. autoclass:: tamm.layers.SecondaryPositionalEncodings
    :members:


Helper function
---------------

.. autofunction:: tamm.layers.transformer.positional_encoding.compute_alibi_slopes
"""

import abc as _abc
import collections as _collections
import math as _math
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm.layers import positional_encoding as _positional_encoding
from tamm.layers import rope as _rope
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import token_metadata as _token_metadata
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalDeviceOrString as _OptionalDeviceOrString

TransformerPositionalEncodingOutput = _collections.namedtuple(
    "TransformerPositionalEncodingOutput", ["embeddings", "attention_side_inputs"]
)
"""
A :class:`namedtuple` for holding outputs from a :class:`TransformerPositionalEncoding`.

.. py:attribute:: embeddings

    Embedding inputs to the first layer of a transformer.

.. py:attribute:: attention_side_inputs

    Side inputs to self attention layers of a transformer.
"""


class TransformerPositionalEncoding(_nn.Module, _abc.ABC):
    """
    A base class for layers that encode token positions into the inputs of
    a transformer layer sequence.
    """

    def __init__(self):
        super().__init__()

    @_abc.abstractmethod
    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        """
        Args:
            token_metadata (:obj:`TokenMetadata`): A :obj:`TokenMetadata`
                that includes an integer ``q_positions`` tensor with shape
                ``(batch_size, seq_len)`` and a ``kv_positions`` tensor with
                shape ``(batch_sie, kv_seq_len)``.  The KV sequence length may be larger
                than the query length when using a KV cache.
            embeddings (:obj:`torch.Tensor`): The input embeddings to a transformer
                model with shape ``(batch_size, seq_len, hidden_dim)``.  The dtype
                should match the computation dtype for attention layers.
            attention_side_inputs (:obj:`dict`): A dictionary of side inputs to
                attention layers.  This should contain ``"attention_mask"`` and
                ``"flash_attention_options"`` keys.

        Returns:
            A :obj:`TransformerPositionalEncodingOutput`, which contains new
            ``embeddings`` and ``attention_side_inputs`` objects with positions encoded.
        """


class NoPositionalEncoding(TransformerPositionalEncoding, _LayerMixin):
    """
    A passthrough NoPE layer, which returns unmodified embeddings and attention
    side inputs.
    """

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        return TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )


class TransformerAbsolutePositionalEmbedding(
    _positional_encoding.AbsolutePositionalEmbedding,
    TransformerPositionalEncoding,
    _ConfigurableLayerMixin,
):
    """
    Layer for absolute positional embeddings.  This layer maps positions to trainable
    embedding tensors and then adds these embeddings to the token embeddings.

    Args:
        embedding: An embedding layer (or builder) that maps positions to embeddings.
        sequence_start (:obj:`int`, optional): An option that controls which tokens
            receive positional embeddings.  If ``sequence_start > 0``, then the first
            ``sequence_start`` tokens do not receive positional embeddings.

    .. note::
        This layer is the same as :class:`AbsolutePositionalEmbedding` except that it
        adapts the :meth:`forward` signature to work with :class:`.TransformerStack`.
    """

    def forward(  # pylint: disable=arguments-differ
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        embeddings = super().forward(embeddings, positions=token_metadata.q_positions)
        return TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )


class SlidingWindowPositionalEncoding(TransformerPositionalEncoding, _LayerMixin):
    """
    Positional encoding layer for sliding window attention
    (`paper <https://arxiv.org/pdf/2004.05150v2>`__).
    This layer masks out attention between tokens when the distance between tokens
    exceeds a window size.

    Args:
        window_size (:obj:int): Number of tokens to attend to on each side.
    """

    def __init__(self, window_size: int):
        """
        Initializes the SlidingWindowPositionalEncoding module.

        Args:
            window_size (:obj:int): Number of tokens in the sliding window on each side.
        """
        super().__init__()
        self.window_size = window_size

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        positions, kv_positions = (
            token_metadata.q_positions,
            token_metadata.kv_positions,
        )

        # Extract positions and compute distances
        distances = _torch.abs(positions[:, :, None] - kv_positions[:, None, :])

        # Create a binary mask based on window size
        distance_threshold = self.window_size
        mask = attention_side_inputs["attention_mask"]
        fill_value = float("-inf") if mask.is_floating_point() else 0
        attention_side_inputs["attention_mask"] = mask.masked_fill(
            distances > distance_threshold, fill_value
        )

        flash_options = attention_side_inputs.get("flash_attention_options")
        if flash_options is not None:
            if "window_size" in flash_options:
                raise RuntimeError("flash_options already contains a window_size")
            flash_options["window_size"] = (self.window_size, self.window_size)

        return TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )

    def extra_repr(self) -> str:
        """
        Returns a string representation of the module's configuration for debugging.

        Returns:
            str: The window size configuration.
        """
        return f"window_size={self.window_size}"


class ALiBiPositionalEncoding(TransformerPositionalEncoding, _LayerMixin):
    """
    Positional Encoding layer for Attention with Linear Biases
    (`paper <https://arxiv.org/abs/2108.12409>`__).
    This layer introduces attention biases that are proportional to the negative
    distance between each query and key.  These distances are scaled by a different
    slope for each attention head.

    Args:
        num_heads (:obj:`int`): The number of attention heads.
    """

    def __init__(self, num_heads: int):
        super().__init__()
        self.num_heads = num_heads

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        positions, kv_positions = (
            token_metadata.q_positions,
            token_metadata.kv_positions,
        )

        mask = attention_side_inputs.get("attention_mask")
        if mask.ndim == 3:
            mask = mask.unsqueeze(-3)  # add heads dim

        biases = positions[:, None, :, None] - kv_positions[:, None, None, :]
        biases = -_torch.abs(biases.type(embeddings.dtype))
        biases = biases.masked_fill(mask.logical_not(), float("-inf"))
        slopes = compute_alibi_slopes(
            self.num_heads, device=biases.device, dtype=biases.dtype
        )
        mask = biases * slopes[None, :, None, None]
        attention_side_inputs["attention_mask"] = mask

        flash_options = attention_side_inputs.get("flash_attention_options")
        if flash_options is not None:
            flash_options["alibi_slopes"] = slopes.type(_torch.float32)

        return TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )

    def extra_repr(self) -> str:
        return f"num_heads={self.num_heads}"


def compute_alibi_slopes(
    num_heads: int,
    *,
    device: _OptionalDeviceOrString = None,
    dtype: _Optional[_torch.dtype] = None,
) -> _torch.Tensor:
    """
    Computes slopes according to the `ALiBi paper <https://arxiv.org/abs/2108.12409>`__.
    Each slope is a scale factor for the ALiBi attention biases corresponding to its
    head.

    Args:
        num_heads (:obj:`int`): The number of heads for attention.
        device (:obj:`str` or :obj:`torch.device`): The target device for the resulting
            slopes tensor.
        dtype (:obj:`torch.dtype`): The dtype for the resulting slopes tensor.

    Returns:
        A 1D :obj:`torch.Tensor` of slopes with length ``num_heads``.
    """
    n = 2 ** _math.floor(_math.log2(num_heads))  # first power of 2 <= num_heads
    first_n = (2 ** (-8 * (i + 1) / n) for i in range(n))
    remainder = (2 ** (-4 * (2 * i + 1) / n) for i in range(num_heads - n))
    slopes = (*first_n, *remainder)
    return _torch.tensor(slopes, device=device, dtype=dtype)


class RotaryPositionalEmbedding(TransformerPositionalEncoding, _LayerMixin):
    """
    Positional encoder layer for Rotary Positional Embeddings
    (`paper <https://arxiv.org/pdf/2104.09864>`__).  This layer computes RoPE rotation
    matrices and adds them to ``attention_side_inputs`` with keys
    ``"query_rope_coefficients"`` and ``"key_rope_coefficients"``.

    Args:
        dim_per_head (:obj:`int`): The hidden dimension per attention head.
        theta (:obj:`float`, optional): ``theta`` param of RoPE.
    """

    def __init__(
        self,
        dim_per_head: int,
        *,
        theta: float = 10000.0,
    ):
        super().__init__()
        self.dim_per_head = dim_per_head
        self.theta = theta

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        coef = _rope.compute_rope_coefficients(
            positions=token_metadata.q_positions,
            dim=self.dim_per_head,
            theta=self.theta,
            dtype=embeddings.dtype,
        )
        attention_side_inputs["query_rope_coefficients"] = coef
        attention_side_inputs["key_rope_coefficients"] = coef

        return TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )

    def extra_repr(self) -> str:
        return f"dim_per_head={self.dim_per_head}, theta={self.theta}"


class SinusoidalPositionalEmbedding(TransformerPositionalEncoding, _LayerMixin):
    """
    Positional encoding layer for Sinusoidal Positional Embeddings
    (`paper <https://arxiv.org/abs/1706.03762>`__). This layer computes sinusoidal
    positional encodings and adds them to the input embeddings.

    Args:
        min_timescale (:obj:`float`, optional): The minimum timescale used for the first half of the embedding channels.
                        Default is 1.0.
        max_timescale (:obj:`float`, optional): The maximum timescale used for the second half of the embedding
                        channels. Default is 10000.0.
    """

    def __init__(
        self,
        *,
        min_timescale: _Optional[float] = 1.0,
        max_timescale: _Optional[float] = 10000.0,
    ):
        super().__init__()

        self.min_timescale = min_timescale
        self.max_timescale = max_timescale

    def extra_repr(self) -> str:
        """Returns a string representation of the object for debugging purposes."""
        return f"min_timescale={self.min_timescale}, max_timescale={self.max_timescale}"

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        positions = token_metadata.q_positions
        # Get the embedding dimension directly from the input tensor
        dim = embeddings.shape[-1]
        num_timescales = dim // 2

        # Calculate the logarithmic timescale increment
        log_timescale_increment = _math.log(
            self.max_timescale / self.min_timescale
        ) / max(1, num_timescales - 1)

        # Generate the exponent values
        exponent = [i * -log_timescale_increment for i in range(num_timescales)]

        # Compute the exponential values
        exp_values = [_math.exp(x) for x in exponent]

        # Compute the inverse timescales for each dimension
        inv_timescales = self.min_timescale * _torch.tensor(
            exp_values, dtype=embeddings.dtype, device=positions.device
        )

        # Scale the positions by the inverse timescales
        scaled_time = positions[..., None] * inv_timescales

        # Concatenate sin and cosine of the scaled time to create the positional embeddings
        pos_embeddings = _torch.cat(
            [_torch.sin(scaled_time), _torch.cos(scaled_time)], dim=-1
        )

        # Add the positional embeddings to the input embeddings
        embeddings = embeddings + pos_embeddings

        return TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )


class SequentialPositionalEncoding(_nn.Module, _LayerMixin):
    """
    A layer that computes multiple positional encodings in sequence.

    Args:
        named_layers (:obj:`dict`): A dictionary that maps the name of each positional encoding
            layer to a :obj:`.LayerBuilder` or :obj:`nn.Module`.
    """

    def __init__(
        self,
        named_layers: _Dict[str, _nn.Module],
    ):
        super().__init__()

        # Initialize and add named layers
        for layer_name, layer in list(named_layers.items()):
            if layer is not None:
                layer = _helpers.maybe_build_module(layer)
                self.add_module(layer_name, layer)
            else:
                setattr(self, layer_name, None)
                named_layers.pop(layer_name)

    @property
    def named_layers(self) -> _collections.abc.ItemsView:
        """The names and layers in the sequence as an :obj:`ItemsView`."""
        return self._modules.items()

    @property
    def layer_names(self) -> _collections.abc.KeysView:
        """The names of layers in the sequence as a :obj:`KeysView`."""
        return self._modules.keys()

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> TransformerPositionalEncodingOutput:
        output = TransformerPositionalEncodingOutput(
            embeddings=embeddings, attention_side_inputs=attention_side_inputs
        )

        for _, layer in self.named_layers:
            output = layer(
                token_metadata=token_metadata,
                embeddings=output.embeddings,
                attention_side_inputs=output.attention_side_inputs,
            )

        return output


class SecondaryPositionalEncodings(_nn.Module, _LayerMixin):
    """
    A layer that computes additional positional encodings in a
    :class:`.TransformerStack` (other than the stack's regular
    ``positional_encoding``).  This is helpful for models that mix
    positional encoding types (such as local and global attention).

    Args:
        named_layers (:obj:`dict`): A dictionary that maps child layer
            names to positional encoding layers.
    """

    def __init__(self, named_layers: _Dict[str, _ModuleOrBuilder]):
        super().__init__()
        _helpers.append_children(self, **named_layers)

    def forward(
        self,
        *,
        token_metadata: _token_metadata.TokenMetadata,
        embeddings: _torch.Tensor,
        attention_side_inputs: _Dict[_Any, _Any],
    ) -> _Dict[str, _Dict[str, _Any]]:
        """
        Arguments are the same as :meth:`.TransformerPositionalEncoding.forward`.

        Returns:
            A dictionary that maps the name of each child layer to its
            ``attention_side_inputs`` output.
        """

        result = {}
        for name, child in self.named_children():
            attention_side_inputs_copy = dict(attention_side_inputs.items())
            _, new_attention_side_inputs = child(
                token_metadata=token_metadata,
                embeddings=embeddings,
                attention_side_inputs=attention_side_inputs_copy,
            )
            result[name] = new_attention_side_inputs
        return result



import enum as _enum
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List

import torch.nn as _nn

from tamm import _helpers
from tamm._helpers import case_insensitive_lookup
from tamm.utils import partial as _partial
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin


class AdapterMode(str, _enum.Enum):
    TRANSFORM_INPUTS = "TRANSFORM_INPUTS"
    TRANSFORM_OUTPUTS = "TRANSFORM_OUTPUTS"

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


class LayerAdapter(_nn.Module):
    """
    Base class for implementing a layer adapter.
    """

    def __init__(self):
        super().__init__()
        self._child_adapters = {}

        self._has_input_transform = (
            _helpers.get_function_from_maybe_method(self._transform_inputs)
            is not LayerAdapter._transform_inputs
            # we set this attribute during __init__ so that we can use
            # :attr:`has_input_transform` during forward() -- if we dynamically
            # compute this instead during :meth:`forward()` it conflicts
            # with torch.compile() and torch.jit.trace()
        )
        self._has_output_transform = (
            _helpers.get_function_from_maybe_method(self._transform_outputs)
            is not LayerAdapter._transform_outputs
        )

    def _transform_inputs(self, *, args: _Any, kwargs: _Any) -> _Any:
        return args, kwargs

    def _transform_outputs(  # pylint: disable=unused-argument
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _Any,
    ) -> _Any:
        return outputs

    @property
    def has_input_transform(self) -> bool:
        """
        Returns :obj:`False` if the adapter uses the base input transform (which is a
        no-op) and :obj:`True` otherwise.
        """
        return self._has_input_transform

    @property
    def has_output_transform(self) -> bool:
        """
        Returns :obj:`False` if the adapter uses the base output transform (which is a
        no-op) and :obj:`True` otherwise.
        """
        return self._has_output_transform

    def forward(
        self,
        mode: AdapterMode,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any = None,
        transformed_kwargs: _Any = None,
        outputs: _Any = None,
    ):
        mode = _helpers.get_enum_member_from_name(AdapterMode, mode)
        if mode is AdapterMode.TRANSFORM_INPUTS:
            return self._transform_inputs(args=args, kwargs=kwargs)
        return self._transform_outputs(
            outputs=outputs,
            args=args,
            kwargs=kwargs,
            transformed_args=transformed_args,
            transformed_kwargs=transformed_kwargs,
        )

    def register_child_adapter(self, name: str, adapter: "LayerAdapter"):
        """
        Register a child adapter.
        """
        self.register_module(name, adapter)
        self._child_adapters[name] = adapter

    def register_child_adapters(self, adapters: _Dict[str, "LayerAdapter"]):
        """
        Register child adapters.
        """
        for ad_name, adapter in adapters.items():
            self.register_child_adapter(ad_name, adapter)

    def freeze(self):
        """
        Make the parameters of the adapter un-trainable.
        """
        for param in self.parameters():
            param.requires_grad = False

    def unfreeze(self):
        """
        Make the parameters of the adapter trainable.
        """
        for param in self.parameters():
            param.requires_grad = True

    def reset_parameters(self):
        """
        Reset adapter parameters.
        """
        for _, adapter in self._child_adapters.items():
            adapter.reset_parameters()

    def reset_adapter_parameters(self):
        """
        Same as :py:meth:`reset_parameters`. Adds backward compatibility with adapters v0 implementation.
        """
        self.reset_parameters()


class MergeableLayerAdapterMixin:
    """
    An adapter mixin which implements a merge method to merge adapter
    weights into the wrapped layer's weights.
    """

    def merge_adapter(self, wrapped_module: _nn.Module):
        raise NotImplementedError("This method is not implemented.")


# pylint: disable-next=abstract-method
class AdapterWithExtraInputs(LayerAdapter):
    """
    An adapter which accepts extra keyword arguments as inputs,
    besides the args and kwargs of the layer it adapts
    """

    extra_input_names: _List[str]


class LayerAdapterConfig(
    _partial.DataclassedPartial,
    _JSONSerializableMixin,
    json_namespace="adapters",
):
    """
    Base class for implementing a layer adapter config.
    """

    def create_adapter(
        self,
        *override_args,
        **override_kwargs,
    ) -> LayerAdapter:
        """
        Create a configured adapter.

        Args:
            override_args: Optional positional arguments to override args specified in
                the config.  These args replace the first ``len(override_args)``
                positional args (and *all* varargs if ``override_args`` contains
                varargs) in the adapter's constructor.
            override_kwargs: Optional keyword override arguments. These arguments
                replace any additional named arguments not overriden by
                ``override_args``.
        Returns:
            The newly created adapter.
        """
        return self(*override_args, **override_kwargs)

    def _to_json_dict_impl(self):
        result = _helpers.dataclass_to_dict(
            self.configured_args,
            omit_defaults=True,  # only include non-defaults for forward compatibility
        )
        return result


def attach_config_class(base_cls):
    """
    A class decorator that creates a config for an adapter
    and attaches that type to the class as the attribute
    ``Config``.
    """
    adapter_name = base_cls.__name__
    config_name = f"{adapter_name}Config"
    config_cls = LayerAdapterConfig.create_subclass(
        target_callable=base_cls, name=config_name
    )
    config_cls.__doc__ = (
        f"A :py:class:`.BaseAdapterConfig` subclass for configuring "
        f":py:class:`.{adapter_name}` adapters.  Use the alias :attr:`."
        f"{adapter_name}.Config` to access this class. "
        f"Please check :class:`.{adapter_name}` for more details about the "
        "signature."
    )
    base_cls.Config = config_cls
    return base_cls


def _validate_adapters(adapters: _Dict[str, LayerAdapter]):
    if not adapters:
        raise ValueError(
            "The input adapters list is empty. In order to construct a composite "
            "adapter, there should be at least one adapter in the list."
        )


class CompositeInputTransform(LayerAdapter):
    """
    A base class for composite layer adapters that only transform the wrapped layer's
    inputs, not outputs.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`LayerAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.
    """

    def __init__(self, adapters: _Dict[str, LayerAdapter]):
        super().__init__()
        _validate_adapters(adapters)
        for adapter in adapters.values():
            if adapter.has_output_transform:
                raise ValueError(
                    "CompositeInputTransform received a child adapter that transforms "
                    "outputs."
                )
        self.register_child_adapters(adapters)


class CompositeOutputTransform(LayerAdapter):
    """
    A base class for composite layer adapters that only transform the wrapped layer's
    outputs, not inputs.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`LayerAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.
    """

    def __init__(self, adapters: _Dict[str, LayerAdapter]):
        super().__init__()
        _validate_adapters(adapters)
        for adapter in adapters.values():
            if adapter.has_input_transform:
                raise ValueError(
                    "CompositeOutputTransform received a child adapter that transforms "
                    "inputs."
                )
        self.register_child_adapters(adapters)



from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch as _torch
import torch.nn as _nn

from tamm.ao.layers.kv_quantizer import KVQuantizer as _KVQuantizer
from tamm.context_vars import get_first_token_generation_flag
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.decoding import KVCacher as _KVCacher
from tamm.layers.transformer import kv_cache as _kv_cache


class QuantizingKVCacher(_nn.Module, _LayerMixin):
    """
    Enhanced KV cacher with quantization support.

    1. Quantizes the key and value before storing them in the kv_cache.
    2. For the first token only, overwrites the cached quantized values with its original
        unquantized tensors, to maintain high-precision.
    """

    def __init__(
        self,
        key_quantizer,
        value_quantizer,
        kv_cacher: _KVCacher,
        freeze_qparams: _Optional[bool] = False,
    ):
        super().__init__()
        self.kv_cacher = kv_cacher
        self.freeze_qparams = freeze_qparams
        self.quantizer = _KVQuantizer(
            key_quantizer=key_quantizer,
            value_quantizer=value_quantizer,
        )

    def forward(
        self,
        query: _torch.Tensor,
        key: _torch.Tensor,
        value: _torch.Tensor,
        kv_cache: _Optional[_kv_cache.KVCacheLayerView] = None,
    ) -> _Tuple[_torch.Tensor, _torch.Tensor, _torch.Tensor]:
        """
        Applies quantization to keys and values before storing in KV-cache, but overwrites with
            unquantized tensors for the first token only.

        Args:
            query (:obj:`torch.Tensor`): batch x L x dim
            key (:obj:`torch.Tensor`): batch x L x dim
            value (:obj:`torch.Tensor`): batch x L x dim
            kv_cache (:obj: `Optional[torch.Tensor]`): 2 x batch x S x dim, where S >= L

        Returns:
            Tuple of tensors [batch x L x dim, batch x S x dim, batch x S x dim]
        """
        query, quant_key, quant_value = self.quantizer(query, key, value)

        query, full_key, full_value = self.kv_cacher(
            query, quant_key, quant_value, kv_cache
        )

        # For first token only, overwrite with unquantized values
        if kv_cache is not None and get_first_token_generation_flag():
            full_key = self._overwrite_tail_entries(full_key, key)
            full_value = self._overwrite_tail_entries(full_value, value)

        return query, full_key, full_value

    @staticmethod
    def _overwrite_tail_entries(
        original_tensor: _torch.Tensor, new_tensor: _torch.Tensor
    ) -> _torch.Tensor:
        """
        Creates a new tensor by concatenating the head of the original tensor with the new tensor.
            Supports both regular models with shape (batch_size, seq_len, hidden_dim) and
            PTT models with shape (num_tracks, batch_size, seq_len, hidden_dim).

        Args:
            original_tensor (:obj:`_torch.Tensor`): The full cached tensor (... x seq_len x dim).
            new_tensor (:obj:`_torch.Tensor`): The new tensor to overwrite (... x new_seq_len x dim).
                Here, new_seq_len <= seq_len.

        Returns:
            :obj:`_torch.Tensor`: A new tensor combining the head of ``original_tensor``
                with ``new_tensor``, maintaining original sequence length.
        """
        if new_tensor.size(-2) > original_tensor.size(-2):
            raise ValueError(
                f"New key/ value tensor sequence length: {new_tensor.size(-2)} "
                f"exceeds original cache size: {original_tensor.size(-2)}"
            )
        tail_start = original_tensor.size(-2) - new_tensor.size(-2)
        return _torch.cat([original_tensor[..., :tail_start, :], new_tensor], dim=-2)



"""
layers.attention
^^^^^^^^^^^^^^^^

This module implements attention layers for transformer models.

.. autoclass:: tamm.layers.attention.SDPAImplementationChoice
    :members:
"""
# pylint: disable=too-many-lines

import enum as _enum
import logging as _logging
import math as _math
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch import _dynamo
from torch import nn as _nn

from tamm import _adapters_v1, _helpers
from tamm._helpers import case_insensitive_lookup
from tamm.layers import decoding as _decoding
from tamm.layers import dropout as _dropout
from tamm.layers import functional as _tamm_F
from tamm.layers import linear as _linear
from tamm.layers import norm as _norm
from tamm.layers import residual as _residual
from tamm.layers import rope as _rope
from tamm.layers import sequential as _sequential
from tamm.layers import side_outputs as _side_outputs
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.flash_attention import _qkv_flash_attention_compatible
from tamm.layers.flash_attention import (
    flash_attn_func_with_low_rank_mask as _flash_attn_func_with_low_rank_mask,
)
from tamm.layers.flash_attention import (
    is_flash_attention_available as _is_flash_attn_available,
)
from tamm.typing import ModuleBuilder as _ModuleBuilder
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder

try:
    # pylint: disable-next=ungrouped-imports
    from torch.nn.attention import flex_attention as _flex_attention  # noqa

    _IS_FLEX_ATTENTION_AVAILABLE = True
except ImportError:
    _IS_FLEX_ATTENTION_AVAILABLE = False


_logger = _logging.getLogger(__name__)


class TransformerAttention(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    Pre-norm attention layer with residual connection.  Takes inputs of shape
    ``(batch_size, sequence_length, target_dim)`` and optional source inputs of shape
    ``(batch_size, sequence_length, source_dim)``.  The output is the result of the
    following layer sequence:

    * Norm
    * Query/key/value transform
    * RoPE transform
    * QK norm
    * KV cacher
    * Scaled dot product attention
    * Output transform
    * Dropout
    * Residual add

    The QKV transform layer should take an optional ``source`` argument, and the KV
    cacher should accept an optional ``kv_cache`` argument.  Also, the SDPA layer should
    take an optional ``attention_mask`` argument and an
    optional ``flash_attention_options`` argument.

    Args:
        norm: The norm layer (:obj:`nn.Module`) or builder (callable that returns the
            :obj:`nn.Module` when called without args).
        qkv_transform: The QKV transform layer or builder.
        rope_transform: The RoPE transform layer or builder.
        qk_norm: Layer for applying scaling on key and query vectors.
            It is optional and defined if QK normalization is applied.
        kv_cacher: The KV cacher layer or builder.
        scaled_dot_product_attention: The SDPA layer or builder.
        output_transform: The output transform layer or builder.
        dropout: The dropout layer or builder.
        residual_connection: The residual connection layer or builder.

    Users can also create this layer using its :meth:`create` method, which has
    simpler args.  Alternatively, use :meth:`create_builder` to create the layer via
    a configurable builder.
    """

    def __init__(
        self,
        *,
        norm: _OptionalModuleOrBuilder,
        qkv_transform: _ModuleOrBuilder,
        rope_transform: _OptionalModuleOrBuilder,
        qk_norm: _OptionalModuleOrBuilder,
        kv_cacher: _ModuleOrBuilder,
        scaled_dot_product_attention: _ModuleOrBuilder,
        output_transform: _ModuleOrBuilder,
        dropout: _ModuleOrBuilder,
        residual_connection: _ModuleOrBuilder,
        kv_dim: float,
        attention_layer_type: _Optional[str] = None,
    ):
        # pylint: disable=too-many-locals
        self.kv_dim = kv_dim

        # We do the following since converters have difficulty
        # distinguishing between layers which have positional encoding
        # which leads to differences in the state dict.
        qkv_transform_key = "qkv_transform"
        if attention_layer_type == "global_nope":
            qkv_transform_key = "qkv_transform_global_nope"

        named_layers = {
            "norm": norm,
            qkv_transform_key: qkv_transform,
            "rope_transform": rope_transform,
            "qk_norm": qk_norm,
            "kv_cacher": kv_cacher,
            "scaled_dot_product_attention": scaled_dot_product_attention,
            "output_transform": output_transform,
            "dropout": dropout,
        }
        side_input_keys = {
            qkv_transform_key: [("source", "key"), ("source", "value")],
            "rope_transform": ["query_rope_coefficients", "key_rope_coefficients"],
            "kv_cacher": ["kv_cache"],
            "scaled_dot_product_attention": [
                "attention_mask",
                "flash_attention_options",
            ],
        }

        if rope_transform is None:
            named_layers.pop("rope_transform")
            side_input_keys.pop("rope_transform")
        if qk_norm is None:
            named_layers.pop("qk_norm")
        super().__init__(
            named_layers,
            side_input_keys=side_input_keys,
            unpack_tuple_inputs=True,
            residual_connection=residual_connection,
        )
        self.qkv_transform_key = qkv_transform_key
        self._mark_adaptable_layers()

    @classmethod
    def create_basic_builder(
        cls,
        *,
        target_dim: int,
        source_dim: _Optional[int] = None,
        num_heads: int = 1,
        num_kv_heads: _Optional[int] = None,
        norm: _Union[str, _OptionalModuleOrBuilder] = "layer_norm",
        apply_rope: bool = False,
        apply_qk_norm: bool = False,
        attention_dropout_p: float = 0.0,
        hidden_dropout_p: float = 0.0,
        qkv_transform_bias: bool = False,
        output_transform_bias: bool = False,
        apply_residual_add: bool = True,
        output_attentions: bool = False,
        output_kv_state: bool = False,
        sdpa_implementation: _Union["SDPAImplementationChoice", str] = "auto",
        vec_dim: _Optional[_List[int]] = None,
        atten_hidden_dim: _Optional[int] = None,
        attention_layer_type: _Optional[str] = None,
        sliding_window_size: _Optional[int] = None,
    ):  # pylint: disable=arguments-differ
        """
        Creates and returns a default builder for creating :obj:`TransformerAttention`
        objects.  The builder uses :class:`LayerNorm` for the norm layer and
        :class:`Linear` for the output transform.  For the QKV transform, the layer uses
        :class:`FusedMultiOutputLinear` if ``source_dim`` is ``None`` and
        :class:`QKVLinear` otherwise.

        Use :meth:`create` to directly create an attention layer.

        Args:
            target_dim (:obj:`int`): The hidden dimension of the target input.  This
                tensor should have shape ``(batch_size, sequence_length, source_dim)``.
            source_dim (:obj:`int`, optional): The hidden dimension of the optional
                source input.
            num_heads (:obj:`int`): The number of attention heads.
            num_kv_heads (:obj:`int`, optional): The number of key-value heads for
                grouped query attention (GQA: https://arxiv.org/pdf/2305.13245.pdf).
                The default value is ``num_heads``, which results in vanilla
                multi-head attention.
            norm: The norm layer (:obj:`nn.Module`) or builder (a callable that returns
                the norm layer when called without args).  If ``None``, this defaults to
                :class:`LayerNorm`.
            apply_rope (:obj:`bool`, optional): Flag for including a RoPE transform
                layer to rotate query and key tensors.  Defaults to ``False``.  If
                ``True``, the attention layer expects ``query_rope_coefficients``
                and ``key_rope_coefficients`` keyword arguments during
                :meth:`.forward`.
            apply_qk_norm (:obj:`bool`, optional): Flag for applying normalization on
                key and query vectors.  Defaults to ``False``.
            attention_dropout_p (:obj:`float`): The dropout probability for attention
                probabilities.
            hidden_dropout_p (:obj:`float`): The dropout probability for the dropout
                layer prior to the residual add.
            qkv_transform_bias (:obj:`bool`): Should the QKV Projection Layer
                use a bias term.
            output_transform_bias (:obj:`bool`): Should the output transform
                linear layer use a bias term.
            apply_residual_add (:obj:`bool`): Should apply a residual add layer.
            output_attentions (:obj:`bool`): Flag that when ``True`` results in the
                inclusion of attentions weights in ScaledDotProductAttention.
            output_kv_state (:obj:`bool`): Flag that when ``True`` results in the
                inclusion of kv_state of current layer as side outputs. Defaults to ``False``.

        Returns:
            A :obj:`TransformerAttentionLayer.Builder` for creating layers.
        """

        # pylint: disable=too-many-locals, duplicate-code

        # Don't apply rope if global_nope (nope = no positional encoding)
        if attention_layer_type == "global_nope":
            apply_rope = False

        norm = _norm.create_norm_builder((target_dim,), spec=norm)

        if target_dim % num_heads != 0:
            raise ValueError(
                "target_dim must be a multiple of num_heads "
                f"(received target_dim={target_dim} and num_heads={num_heads})"
            )

        num_kv_heads = num_kv_heads if num_kv_heads is not None else num_heads
        if num_heads % num_kv_heads != 0:
            raise ValueError(
                "num_heads must be a multiple of num_kv_heads"
                f"(received num_heads={num_heads} and num_kv_heads={num_kv_heads})"
            )

        atten_hidden_dim = (
            atten_hidden_dim if atten_hidden_dim is not None else target_dim
        )
        kv_output_dim = (atten_hidden_dim * num_kv_heads) // num_heads

        qkvlinear_kwargs = {}
        multioutputlinear_kwargs = {}
        linear_kwargs = {}
        linear_cls = _linear.Linear
        if vec_dim is not None:
            qkvlinear_kwargs = {"vec_dim": vec_dim}
            multioutputlinear_kwargs = {"vec_dim": vec_dim}
            linear_kwargs = {"vec_dim": vec_dim}
            linear_cls = _linear.VectorizedLinear

        if source_dim is not None:
            qkv_transform = QKVLinear.Builder(
                query_dim=target_dim,
                key_dim=source_dim,
                value_dim=source_dim,
                key_output_dim=kv_output_dim,
                value_output_dim=kv_output_dim,
                bias=qkv_transform_bias,
                **qkvlinear_kwargs,
            )
        else:
            qkv_transform = _linear.FusedMultiOutputLinear.Builder(
                target_dim,
                [atten_hidden_dim, kv_output_dim, kv_output_dim],
                bias=qkv_transform_bias,
                **multioutputlinear_kwargs,
            )

        rope_transform = RoPETransform.Builder() if apply_rope else None

        qk_norm = None
        if apply_qk_norm:
            dim_per_head = atten_hidden_dim // num_heads
            qk_norm = QKNorm.Builder(
                dim_per_head=dim_per_head,
                query_norm=_norm.LayerNorm.Builder([dim_per_head]),
                key_norm=_norm.LayerNorm.Builder([dim_per_head]),
            )

        scaled_dot_product_attention = ScaledDotProductAttention.Builder(
            num_heads,
            num_kv_heads=num_kv_heads,
            dropout_p=attention_dropout_p,
            output_attentions=output_attentions,
            implementation=sdpa_implementation,
            sliding_window_size=sliding_window_size
            if attention_layer_type == "local"
            else None,
            output_kv_state=output_kv_state,
        )

        output_transform = linear_cls.Builder(
            atten_hidden_dim, target_dim, bias=output_transform_bias, **linear_kwargs
        )

        dropout = _dropout.Dropout.Builder(p=hidden_dropout_p)

        residual_connection = (
            _residual.ResidualAdd.Builder() if apply_residual_add else None
        )

        return cls.Builder(
            norm=norm,
            qkv_transform=qkv_transform,
            rope_transform=rope_transform,
            qk_norm=qk_norm,
            kv_cacher=_decoding.KVCacher.Builder(),
            scaled_dot_product_attention=scaled_dot_product_attention,
            output_transform=output_transform,
            dropout=dropout,
            residual_connection=residual_connection,
            kv_dim=kv_output_dim,
            attention_layer_type=attention_layer_type,
        )

    def _mark_adaptable_layers(self):
        qkv_transform = getattr(self, self.qkv_transform_key)

        output_transform_layer_type = (
            "BatchedAttentionOutputTransform"
            if isinstance(self.output_transform, _linear.VectorizedLinear)
            else "AttentionOutputTransform"
        )
        _adapters_v1.annotate_layer(
            self.output_transform,
            [(output_transform_layer_type,)],
        )
        if isinstance(qkv_transform, _linear.FusedMultiOutputLinear):
            qkv_transform_layer_type = "AttentionInputTransformFused"
            fused_linear_layer_fields = {
                "input_dim_field": "input_dim",
                "output_dims_field": "output_dims",
                "weight_param_name": "fused_linear.weight",
            }
            if isinstance(qkv_transform.fused_linear, _linear.VectorizedLinear):
                qkv_transform_layer_type = "BatchedAttentionInputTransformFused"
                fused_linear_layer_fields["vec_dim_field"] = "vec_dim"
            _adapters_v1.annotate_layer(
                qkv_transform,
                [(qkv_transform_layer_type, fused_linear_layer_fields)],
            )
        elif isinstance(qkv_transform, QKVLinear):
            layer_type = (
                "BatchedAttentionInputTransformQuery"
                if isinstance(qkv_transform.query_linear, _linear.VectorizedLinear)
                else "AttentionInputTransformQuery"
            )
            _adapters_v1.annotate_layer(
                qkv_transform.query_linear,
                [(layer_type,)],
            )
            layer_type = (
                "BatchedAttentionInputTransformKey"
                if isinstance(qkv_transform.key_linear, _linear.VectorizedLinear)
                else "AttentionInputTransformKey"
            )
            _adapters_v1.annotate_layer(
                qkv_transform.key_linear,
                [(layer_type,)],
            )
            layer_type = (
                "BatchedAttentionInputTransformValue"
                if isinstance(qkv_transform.value_linear, _linear.VectorizedLinear)
                else "AttentionInputTransformValue"
            )
            _adapters_v1.annotate_layer(
                qkv_transform.value_linear,
                [(layer_type,)],
            )


class KVReuseTransformerAttention(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    An attention layer that reuses keys and values from an earlier layer. Takes inputs
    of shape ``(batch_size, sequence_length, target_dim)``.  The output is the result
    of the following layer sequence:

    * Norm
    * Query transform
    * RoPE transform
    * Query norm
    * Scaled dot product attention
    * Output transform
    * Dropout
    * Residual add

    The SDPA layer should take in an external key and value, an optional ``attention_mask``
    argument, and an optional ``flash_attention_options`` argument.

    Args:
        norm: The norm layer (:obj:`nn.Module`) or builder (callable that returns the
            :obj:`nn.Module` when called without args).
        q_transform: The query transform layer or builder.
        rope_transform: The RoPE transform layer or builder.
        q_norm: Layer for applying scaling on key and query vectors.
            It is optional and defined if query normalization is applied.
        scaled_dot_product_attention: The SDPA layer or builder.
        output_transform: The output transform layer or builder.
        dropout: The dropout layer or builder.
        residual_connection: The residual connection layer or builder.
    """

    def __init__(
        self,
        *,
        norm: _OptionalModuleOrBuilder,
        q_transform: _ModuleOrBuilder,
        rope_transform: _OptionalModuleOrBuilder,
        q_norm: _OptionalModuleOrBuilder,
        scaled_dot_product_attention: _ModuleOrBuilder,
        output_transform: _ModuleOrBuilder,
        dropout: _ModuleOrBuilder,
        residual_connection: _ModuleOrBuilder,
    ):
        named_layers = {
            "norm": norm,
            "q_transform": q_transform,
            "rope_transform": rope_transform,
            "q_norm": q_norm,
            "scaled_dot_product_attention": scaled_dot_product_attention,
            "output_transform": output_transform,
            "dropout": dropout,
        }
        side_input_keys = {
            "rope_transform": ["query_rope_coefficients"],
            "scaled_dot_product_attention": [
                "key",
                "value",
                "attention_mask",
                "flash_attention_options",
            ],
        }
        if rope_transform is None:
            named_layers.pop("rope_transform")
            side_input_keys.pop("rope_transform")
        if q_norm is None:
            named_layers.pop("q_norm")
        super().__init__(
            named_layers,
            side_input_keys=side_input_keys,
            unpack_tuple_inputs=True,
            residual_connection=residual_connection,
        )
        self._mark_adaptable_layers()

    @classmethod
    def create_basic_builder(
        cls,
        *,
        target_dim: int,
        num_heads: int = 1,
        num_kv_heads: _Optional[int] = None,
        norm: _Union[str, _OptionalModuleOrBuilder] = "layer_norm",
        apply_rope: bool = False,
        apply_q_norm: bool = False,
        attention_dropout_p: float = 0.0,
        hidden_dropout_p: float = 0.0,
        q_transform_bias: bool = False,
        output_transform_bias: bool = False,
        apply_residual_add: bool = True,
        output_attentions: bool = False,
        output_kv_state: bool = False,
        sdpa_implementation: _Union["SDPAImplementationChoice", str] = "auto",
    ):  # pylint: disable=arguments-differ
        """
        Creates and returns a default builder for creating :obj:`KVReuseTransformerAttention`
        objects.  The builder uses :class:`LayerNorm` for the norm layer and
        :class:`Linear` for the output transform. The norm layer, q_transform layer,
        rope_transform layer, q_norm layer only applies to query tensors.

        Use :meth:`create` to directly create an attention layer.

        Args:
            target_dim (:obj:`int`): The hidden dimension of the target input.  This
                tensor should have shape ``(batch_size, sequence_length, source_dim)``.
            num_heads (:obj:`int`): The number of attention heads.
            num_kv_heads (:obj:`int`, optional): The number of key-value heads for
                grouped query attention (GQA: https://arxiv.org/pdf/2305.13245.pdf).
                The default value is ``num_heads``, which results in vanilla
                multi-head attention.
            norm: The norm layer (:obj:`nn.Module`) or builder (a callable that returns
                the norm layer when called without args).  If ``None``, this defaults to
                :class:`LayerNorm`.
            apply_rope (:obj:`bool`, optional): Flag for including a RoPE transform
                layer to rotate only query tensors.  Defaults to ``False``.  If
                ``True``, the attention layer expects ``query_rope_coefficients``
                keyword arguments during :meth:`.forward`.
            apply_q_norm (:obj:`bool`, optional): Flag for applying normalization on
                only query vectors.  Defaults to ``False``.
            attention_dropout_p (:obj:`float`): The dropout probability for attention
                probabilities.
            hidden_dropout_p (:obj:`float`): The dropout probability for the dropout
                layer prior to the residual add.
            q_transform_bias (:obj:`bool`): Should the Query Projection Layer
                use a bias term.
            output_transform_bias (:obj:`bool`): Should the output transform
                linear layer use a bias term.
            apply_residual_add (:obj:`bool`): Should apply a residual add layer.
            output_attentions (:obj:`bool`): Flag that when ``True`` results in the
                inclusion of attentions weights in ScaledDotProductAttention.
            output_kv_state (:obj:`bool`): Flag that when ``True`` results in the
                inclusion of kv_state of current layer as side outputs. Defaults to ``False``.

        Returns:
            A :obj:`KVReuseTransformerAttention.Builder` for creating layers.
        """

        # pylint: disable=too-many-locals, duplicate-code

        norm = _norm.create_norm_builder((target_dim,), spec=norm)

        if target_dim % num_heads != 0:
            raise ValueError(
                "target_dim must be a multiple of num_heads "
                f"(received target_dim={target_dim} and num_heads={num_heads})"
            )

        num_kv_heads = num_kv_heads if num_kv_heads is not None else num_heads
        if num_heads % num_kv_heads != 0:
            raise ValueError(
                "num_heads must be a multiple of num_kv_heads"
                f"(received num_heads={num_heads} and num_kv_heads={num_kv_heads})"
            )

        linear_kwargs = {}
        linear_cls = _linear.Linear

        q_transform = linear_cls.Builder(
            target_dim, target_dim, bias=q_transform_bias, **linear_kwargs
        )

        rope_transform = RoPETransform.Builder() if apply_rope else None

        q_norm = None
        if apply_q_norm:
            dim_per_head = target_dim // num_heads
            q_norm = QKNorm.Builder(
                dim_per_head=dim_per_head,
                query_norm=_norm.LayerNorm.Builder([dim_per_head]),
            )
        scaled_dot_product_attention = ScaledDotProductAttention.Builder(
            num_heads,
            num_kv_heads=num_kv_heads,
            dropout_p=attention_dropout_p,
            output_attentions=output_attentions,
            implementation=sdpa_implementation,
            output_kv_state=output_kv_state,
        )

        output_transform = linear_cls.Builder(
            target_dim, target_dim, bias=output_transform_bias, **linear_kwargs
        )

        dropout = _dropout.Dropout.Builder(p=hidden_dropout_p)

        residual_connection = (
            _residual.ResidualAdd.Builder() if apply_residual_add else None
        )

        return cls.Builder(
            norm=norm,
            q_transform=q_transform,
            rope_transform=rope_transform,
            q_norm=q_norm,
            scaled_dot_product_attention=scaled_dot_product_attention,
            output_transform=output_transform,
            dropout=dropout,
            residual_connection=residual_connection,
        )

    def _mark_adaptable_layers(self):
        _adapters_v1.annotate_layer(
            self.q_transform,
            [("AttentionInputTransformQuery",)],
        )

        _adapters_v1.annotate_layer(
            self.output_transform,
            [("AttentionOutputTransform",)],
        )


def _split_batch_dimensions(output, original_value_shape):
    original_batch_shapes = original_value_shape[0:2]
    rest_shape = output.shape[1:]
    output = output.reshape(original_batch_shapes + rest_shape)

    return output


def _merge_batch_dimensions(query, key, value, bias):
    # Optimization for SDPA:
    #
    # We know that qkv has shape (N,...,L or S,E or Ev) (this may be transposed in non-torch format)
    #
    # In some cases (i.e: with 5d shape, for example, (batch, tracks, heads, seqlen, target_dim),
    # native torch does not use memory efficient attention. Additionally, Flash Attention assume 4d input.
    # So we merge dimensions in the front with batch so that
    # pytorch uses memory-efficient attention and flash attention can be used.
    #
    # Currently this only works with 5-D inputs (i.e: VectorizedParallelTrackTransformer)
    def merged_optional_shape(x):
        return _math.prod(x.shape[0:2])

    optimize_merge_batch_dimensions = (
        (len(query.shape) == 5)
        and (len(query.shape) == len(value.shape) == len(key.shape) == len(bias.shape))
        and (
            merged_optional_shape(query)
            == merged_optional_shape(key)
            == merged_optional_shape(value)
        )
    )

    if optimize_merge_batch_dimensions:
        # We need to repeat the attn mask before reshaping
        # This looks like (batch, 8, 1, 4096, 4096) for PTT
        # Which is a reasonable amnt of memory.
        bias = bias.repeat(1, query.shape[1], 1, 1, 1)

        query = query.flatten(end_dim=1)
        key = key.flatten(end_dim=1)
        value = value.flatten(end_dim=1)
        bias = bias.flatten(end_dim=1)

    return (query, key, value, bias, optimize_merge_batch_dimensions)


def _add_heads_dim(*tensors, dim_per_head):
    """Reshapes the features dimension of tensors to (heads, features per head)"""
    if len(tensors) != 1:
        return tuple(_add_heads_dim(t, dim_per_head=dim_per_head) for t in tensors)
    tensor = tensors[0]
    shape = tensor.shape
    return tensor.view(*shape[:-1], -1, dim_per_head)


def _remove_heads_dim(*tensors):
    if len(tensors) != 1:
        return tuple(_remove_heads_dim(t) for t in tensors)
    tensor = tensors[0]
    return tensor.flatten(start_dim=-2)


class QKVLinear(
    _nn.Module, _LayerMixin
):  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        query_dim: int,
        *,
        key_dim: _Optional[int] = None,
        value_dim: _Optional[int] = None,
        query_output_dim: _Optional[int] = None,
        key_output_dim: _Optional[int] = None,
        value_output_dim: _Optional[int] = None,
        bias: bool = False,
        vec_dim: _Optional[_List[int]] = None,
        device=None,
        dtype=None,
    ):
        super().__init__()

        self.query_dim = query_dim
        self.key_dim = key_dim if key_dim is not None else query_dim
        self.value_dim = value_dim if value_dim is not None else query_dim

        self.query_output_dim = (
            self.query_dim if query_output_dim is None else query_output_dim
        )
        self.key_output_dim = self.key_dim if key_output_dim is None else key_output_dim
        self.value_output_dim = (
            self.value_dim if value_output_dim is None else value_output_dim
        )
        dtype = _helpers.get_dtype_from_maybe_string(dtype)
        tensor_meta = {"device": device, "dtype": dtype}

        # Handle vectorized
        self.vec_dim = vec_dim
        if vec_dim is not None:
            linear_cls = _linear.VectorizedLinear
            tensor_meta["vec_dim"] = vec_dim
        else:
            linear_cls = _linear.Linear

        self.query_linear = linear_cls(
            self.query_dim, self.query_output_dim, bias=bias, **tensor_meta
        )
        self.key_linear = linear_cls(
            self.key_dim, self.key_output_dim, bias=bias, **tensor_meta
        )
        self.value_linear = linear_cls(
            self.value_dim, self.value_output_dim, bias=bias, **tensor_meta
        )

    def forward(self, query, key=None, value=None):
        key = key if key is not None else query
        value = value if value is not None else query
        return (
            self.query_linear(query),
            self.key_linear(key),
            self.value_linear(value),
        )


class RoPETransform(_nn.Module, _LayerMixin):
    def __init__(self):
        super().__init__()

    def forward(
        self,
        query: _torch.Tensor,
        key: _Optional[_torch.Tensor] = None,
        value: _Optional[_torch.Tensor] = None,
        *,
        query_rope_coefficients: _torch.Tensor,
        key_rope_coefficients: _Optional[_torch.Tensor] = None,
    ):
        dim_per_head = 2 * query_rope_coefficients.size(-3)
        if key is not None and key_rope_coefficients is None:
            raise ValueError(
                "key_rope_coefficients should not be None when key is not None since key transform is expected."
            )
        if key is None:
            query = _add_heads_dim(query, dim_per_head=dim_per_head)
            query = _rope.rotate(query, rope_coefficients=query_rope_coefficients)
            query = _remove_heads_dim(query)
            return query

        query, key = _add_heads_dim(query, key, dim_per_head=dim_per_head)
        query = _rope.rotate(query, rope_coefficients=query_rope_coefficients)
        key = _rope.rotate(key, rope_coefficients=key_rope_coefficients)
        query, key = _remove_heads_dim(query, key)
        return query, key, value


class QKNorm(_nn.Module, _LayerMixin):
    def __init__(
        self,
        dim_per_head: int,
        *,
        query_norm: _Optional[_ModuleBuilder] = None,
        key_norm: _Optional[_ModuleBuilder] = None,
    ):
        super().__init__()
        self.dim_per_head = dim_per_head
        self.query_norm = _helpers.maybe_build_module(query_norm)
        self.key_norm = _helpers.maybe_build_module(key_norm)

    def forward(
        self,
        query: _torch.Tensor,
        key: _Optional[_torch.Tensor] = None,
        value: _Optional[_torch.Tensor] = None,
    ) -> _torch.Tensor:
        """Scales the projected queries and keys."""
        if key is None:
            query = _add_heads_dim(query, dim_per_head=self.dim_per_head)
        else:
            query, key = _add_heads_dim(query, key, dim_per_head=self.dim_per_head)

        if self.query_norm is not None:
            query = self.query_norm(query)
        if self.key_norm is not None:
            key = self.key_norm(key)

        if key is None:
            query = _remove_heads_dim(query)
            return query
        query, key = _remove_heads_dim(query, key)
        return query, key, value


class SDPAImplementationChoice(str, _enum.Enum):
    """
    An :class:`Enum` for specifying the implementation of scaled dot product attention.
    """

    #: Delegate the implementation choice to |tamm|, which attempts to select the most
    #: efficient available implementation
    AUTO = "AUTO"

    #: Use ``torch.nn.functional.scaled_dot_product_attention()``
    NATIVE_TORCH = "NATIVE_TORCH"

    #: Use :func:`flash_attn.flash_attn_func`
    FLASH = "FLASH"

    #: Use :func:`torch.nn.attention.flex_attention()`
    FLEX = "FLEX"

    #: Use ``xformers.ops.memory_efficient_attention()``
    XFORMERS_MEMORY_EFFICIENT = "XFORMERS_MEMORY_EFFICIENT"

    #: Use a basic PyTorch implementation, which can also output attention probabilities
    BASIC = "BASIC"

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


class ScaledDotProductAttention(_nn.Module, _LayerMixin):
    """
    Multihead scaled dot product attention layer.  Takes inputs ``query``, ``key``,
    and ``value``.  The shape for ``query`` is
    ``(batch_size, sequence_length, num_heads * dim_per_head)``, and the shape for
    ``key`` and ``value`` is
    ``(batch_size, sequence_length, num_kv_heads * dim_per_head)``.
    Returns an output with the same shape as ``query``.

    Args:
        num_heads (:obj:`int`): The number of attention heads.
        num_kv_heads (:obj:`int`, optional): The number of key-value heads for grouped
            query attention.  This is ``num_heads`` by default, which results in vanilla
            multi-head attention.
        dropout_p (:obj:`float`): The dropout probability.
        implementation (:obj:`SDPAImplementationChoice` member or :obj:`str`): The
            attention implementation to use.  The layer uses ``AUTO`` by default, which
            attempts to select the most efficient implementation available.  If a
            :obj:`str`, this argument must equal the name of a
            :class:`.SDPAImplementationChoice` member (case insensitive).
        scale (:obj:`float`, optional): Scaling factor prior to the softmax.  Defaults
            to ``1 / sqrt(dim_per_head)``.
        output_attentions (:obj:`bool`): Flag that when ``True`` results in the
            inclusion of attentions weights as side outputs. Defaults to ``False``.
        output_kv_state (:obj:`bool`): Flag that when ``True`` results in the
            inclusion of kv_state of current layer as side outputs. Defaults to ``False``.
    """

    def __init__(
        self,
        num_heads: int,
        *,
        num_kv_heads: _Optional[int] = None,
        scale: _Optional[float] = None,
        logits_soft_cap: _Optional[float] = None,
        dropout_p: float = 0.0,
        implementation: _Union[SDPAImplementationChoice, str] = "auto",
        output_attentions: bool = False,
        output_kv_state: bool = False,
        sliding_window_size: _Optional[int] = None,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.num_kv_heads = num_kv_heads if num_kv_heads is not None else num_heads
        self.sliding_window_size = sliding_window_size
        if self.num_heads % self.num_kv_heads > 0:
            raise ValueError(
                f"Num kv heads ({self.num_kv_heads}) does not evenly divide num heads "
                f"({self.num_heads})"
            )

        self.logits_soft_cap = logits_soft_cap
        self.dropout_p = dropout_p
        self.implementation = _helpers.get_enum_member_from_name(
            SDPAImplementationChoice, implementation
        )
        self.scale = scale
        self.output_attentions = output_attentions
        self.output_kv_state = output_kv_state

        # Call this function at initialization time to make sure the result is cached
        # so that `torch.compile` works
        _is_flash_attn_available()

        is_impl_flash = self.implementation is SDPAImplementationChoice.FLASH
        if is_impl_flash and not _is_flash_attn_available():
            raise RuntimeError(
                "Created an SDPA layer with flash attention as the implementation "
                "choice, but flash_attn is not available."
            )

    # pylint: disable-next=all
    def forward(
        self,
        query: _torch.Tensor,
        key: _torch.Tensor,
        value: _torch.Tensor,
        attention_mask: _Optional[_torch.Tensor] = None,
        flash_attention_options: _Optional[_Dict] = None,
    ):
        """
        Args:
            query (:obj:`torch.Tensor`): Query tensor of shape ``(batch_size, seq_len_q, num_heads * dim_per_head)``.
            key (:obj:`torch.Tensor`): Key tensor of shape ``(batch_size, seq_len_kv, num_kv_heads * dim_per_head)``.
            value (:obj:`torch.Tensor`): Value tensor of shape ``(batch_size, seq_len_kv, num_kv_heads * dim_per_head)``
            attention_mask (:obj:`torch.Tensor` or Optional): Mask for self-attention or cross-attention of shape
                ``(batch_size, seq_len_q, seq_len_kv)``.
            flash_attention_options (:obj:`dict` or Optional): Additional options for flash attention.
        """

        if self.output_kv_state:
            input_key, input_value = key, value
        dim_per_head = query.size(-1) // self.num_heads
        query, key, value = _add_heads_dim(query, key, value, dim_per_head=dim_per_head)

        if attention_mask is not None and attention_mask.ndim == 3:
            attention_mask = attention_mask[:, None, ...]

        implementation = self._get_implementation(
            query, key, value, flash_options=flash_attention_options
        )

        original_value_shape = value.shape
        (
            query,
            key,
            value,
            attention_mask,
            did_merge_batch_dimension,
        ) = _merge_batch_dimensions(query, key, value, attention_mask)

        if (
            self.output_attentions
            and implementation is not SDPAImplementationChoice.BASIC
        ):
            raise ValueError(
                "Only `BASIC` implementation supports output_attentions. "
                "Please set the implementation to either `AUTO` or `BASIC`."
            )

        kwargs = {
            "query": query,
            "key": key,
            "value": value,
            "bias": attention_mask,
            "scale": self.scale,
            "logits_soft_cap": self.logits_soft_cap,
            "dropout_p": self.dropout_p if self.training else 0.0,
        }

        if implementation is SDPAImplementationChoice.NATIVE_TORCH:
            output = self._sdpa_native_torch(**kwargs)
        elif implementation is SDPAImplementationChoice.FLEX:
            output = self._sdpa_flex(**kwargs)
        elif implementation is SDPAImplementationChoice.XFORMERS_MEMORY_EFFICIENT:
            output = self._sdpa_xformers_memory_efficient(**kwargs)
        elif implementation is SDPAImplementationChoice.BASIC:
            output = self._sdpa_basic(**kwargs)
        elif implementation is SDPAImplementationChoice.FLASH:
            if flash_attention_options is None:
                raise ValueError(
                    "Using Flash SDPA implementation but forward() received None for "
                    "flash_attention_options"
                )
            kwargs = {
                key: val
                for key, val in kwargs.items()
                if key not in ("bias", "logits_soft_cap")
            }
            kwargs.update(flash_attention_options)

            # If merged batch dimension (i.e: in case of parallel track models)
            # we need to repeat some of the arguments for flash attention
            # so that the first dimension matches.
            if did_merge_batch_dimension:
                for flash_attn_key_to_update in [
                    "query_bias",
                    "key_bias",
                    "value_padding",
                ]:
                    if flash_attn_key_to_update not in kwargs:
                        continue
                    tensor = kwargs[flash_attn_key_to_update]
                    repeat_times = query.size(0) // tensor.size(0)
                    tensor = tensor.repeat_interleave(repeat_times, dim=0)
                    kwargs[flash_attn_key_to_update] = tensor

            if self.sliding_window_size is not None:
                kwargs["window_size"] = [self.sliding_window_size, 0]

            if self.logits_soft_cap is not None:
                raise NotImplementedError(
                    "logits_soft_cap is currently unsupported for flash attention"
                )

            output = _flash_attn_func_with_low_rank_mask(**kwargs)
        else:
            raise ValueError(f"SDPA implementation {implementation} not recognized")

        if isinstance(output, _side_outputs.OutputWithSideOutputs):
            new_output = output.output
            if did_merge_batch_dimension:
                new_output = _split_batch_dimensions(
                    output.output, original_value_shape
                )
            new_output = _remove_heads_dim(new_output)
            side_outputs = output.side_outputs
            if self.output_kv_state:
                side_outputs = _side_outputs.merge_side_outputs(
                    side_outputs, {"key": input_key, "value": input_value}
                )
            return _side_outputs.OutputWithSideOutputs(
                new_output, side_outputs=side_outputs
            )

        if did_merge_batch_dimension:
            output = _split_batch_dimensions(output, original_value_shape)
        output = _remove_heads_dim(output)
        if self.output_kv_state:
            output = _side_outputs.OutputWithSideOutputs(
                output, side_outputs={"key": input_key, "value": input_value}
            )
        return output

    def _get_implementation(self, query, key, value, *, flash_options):
        if self.implementation is SDPAImplementationChoice.AUTO:
            return self._get_auto_implementation(
                query, key, value, flash_options=flash_options
            )
        return self.implementation

    def _get_auto_implementation(self, query, key, value, *, flash_options):
        if self.output_attentions:
            return SDPAImplementationChoice.BASIC
        if self.logits_soft_cap is not None:
            return SDPAImplementationChoice.BASIC

        if (
            flash_options is not None
            and _is_flash_attn_available()
            and _qkv_flash_attention_compatible(query, key, value)
            and self.num_heads <= 256
            and self.num_kv_heads <= 256
        ):
            return SDPAImplementationChoice.FLASH

        return SDPAImplementationChoice.NATIVE_TORCH

    def _sdpa_xformers_memory_efficient(
        self,
        *,
        query,
        key,
        value,
        bias,
        dropout_p,
        scale=None,
        logits_soft_cap=None,
    ):
        if logits_soft_cap is not None:
            raise NotImplementedError(
                "XFORMERS_MEMORY_EFFICIENT attention does not support logits_soft_cap"
            )
        query, key, value = self._maybe_tile_qkv_for_gqa(query, key, value)
        output = _xformers_optimized_attention(
            query,
            key,
            value,
            attn_bias=bias,
            p=dropout_p,
            scale=scale,
        )
        return output

    def _sdpa_native_torch(
        self, *, query, key, value, bias, dropout_p, scale=None, logits_soft_cap=None
    ):
        if logits_soft_cap is not None:
            raise NotImplementedError(
                "NATIVE_TORCH attention does not support logits_soft_cap"
            )

        query, key, value = self._maybe_tile_qkv_for_gqa(query, key, value)
        query = query.transpose(-3, -2)
        key = key.transpose(-3, -2)
        value = value.transpose(-3, -2)

        if self.sliding_window_size is not None:
            bias = bias.triu_(-self.sliding_window_size)

        kwargs = {"attn_mask": bias, "dropout_p": dropout_p, "is_causal": False}
        if scale is not None:
            # don't include this always because it's not supported in torch 2.0
            kwargs["scale"] = scale
        # pylint: disable-next=not-callable
        output = _nn.functional.scaled_dot_product_attention(
            query, key, value, **kwargs
        )

        return output.transpose(-3, -2)

    def _sdpa_basic(
        self, query, key, value, bias, dropout_p, scale=None, logits_soft_cap=None
    ):
        """
        Returns attentions weights as side outputs,
        which is calculated after the attention softmax.
        Attentions weights are used to compute the weighted average
        in the self-attention heads. This method is used when
        output_attentions is set to ``True``.
        Ref: torch.nn.functional.scaled_dot_product_attention
        """
        query, key, value = self._maybe_tile_qkv_for_gqa(query, key, value)
        query = query.transpose(-3, -2)
        key = key.transpose(-3, -2)
        value = value.transpose(-3, -2)

        scale_factor = 1 / _math.sqrt(query.size(-1)) if scale is None else scale

        attn_weight = (scale_factor * query) @ key.transpose(-2, -1)

        if logits_soft_cap is not None:
            attn_weight = _tamm_F.soft_cap(attn_weight, cap=logits_soft_cap)

        if bias is not None:
            if bias.dtype is _torch.bool:
                zero = _torch.tensor(0.0, dtype=attn_weight.dtype, device=bias.device)
                neginf = _torch.tensor(
                    float("-inf"), dtype=attn_weight.dtype, device=bias.device
                )
                bias = _torch.where(bias, zero, neginf)
            attn_weight += bias

        attn_weight = _torch.softmax(attn_weight, dim=-1)

        dropout_attn_weight = _torch.dropout(
            attn_weight, dropout_p, train=self.training
        )
        attn = dropout_attn_weight @ value
        attn = attn.transpose(-3, -2)
        if self.output_attentions:
            return _side_outputs.OutputWithSideOutputs(
                attn, side_outputs={"attentions": attn_weight}
            )
        return attn

    # pylint: disable=unused-argument,too-many-locals
    def _sdpa_flex(
        self, *, query, key, value, bias, dropout_p, scale=None, logits_soft_cap=None
    ):
        # pylint: disable=import-outside-toplevel
        from torch.nn.attention import flex_attention as _flex_attention  # noqa

        if dropout_p != 0.0:
            raise NotImplementedError("flex attention does not support dropout_p")

        query = query.transpose(-3, -2)
        key = key.transpose(-3, -2)
        value = value.transpose(-3, -2)

        if self.sliding_window_size is not None:
            bias = bias.triu_(-self.sliding_window_size)

        if bias is not None and bias.is_floating_point():
            raise NotImplementedError(
                "FLEX attention does not support floating point attention masks"
            )

        bool_mask = bias
        h_dim = bool_mask.size(1)
        if h_dim == 1:
            bool_mask = bool_mask.expand(-1, query.shape[1], -1, -1)

        def mask_mod(b, h, q_idx, kv_idx):
            return bool_mask[b, h, q_idx, kv_idx]

        block_mask = _flex_attention.create_block_mask(
            mask_mod,
            B=query.shape[0],
            H=query.shape[1],
            Q_LEN=query.shape[2],
            KV_LEN=key.shape[2],
            device=str(query.device),
        )

        kwargs = {"block_mask": block_mask}
        if self.num_heads != self.num_kv_heads:
            kwargs["enable_gqa"] = True

        if logits_soft_cap is not None:

            def compute_soft_cap_score(score, *_):
                return _tamm_F.soft_cap(score, cap=logits_soft_cap)

            kwargs["score_mod"] = compute_soft_cap_score

        if scale is not None:
            kwargs["scale"] = scale

        output = _flex_attention.flex_attention(query, key, value, **kwargs)

        return output.transpose(-3, -2)

    def _maybe_tile_qkv_for_gqa(
        self, query: _torch.Tensor, key: _torch.Tensor, value: _torch.Tensor
    ) -> _Tuple[_torch.Tensor, _torch.Tensor, _torch.Tensor]:
        key = self._tile_single_input_for_gqa(key)
        value = self._tile_single_input_for_gqa(value)
        return query, key, value

    def _tile_single_input_for_gqa(self, key_or_value: _torch.Tensor) -> _torch.Tensor:
        heads_ratio = self.num_heads // self.num_kv_heads
        if heads_ratio == 1:
            return key_or_value
        result = _torch.tile(key_or_value, (heads_ratio,))
        return result.reshape(*result.shape[:-2], self.num_heads, -1)

    def extra_repr(self) -> str:
        implementation = self.implementation.name.lower()
        return (
            f"num_heads={self.num_heads},\n"
            f"scale={self.scale},\n"
            f"logits_soft_cap={self.logits_soft_cap},\n"
            f"dropout_p={self.dropout_p},\n"
            f"implementation='{implementation}',"
        )


@_dynamo.disable
def _xformers_optimized_attention(*args, **kwargs):
    # pylint: disable-next=import-outside-toplevel,import-error
    from xformers.ops import memory_efficient_attention

    return memory_efficient_attention(*args, **kwargs)


# DEPRECATED
# pylint: disable=all
from tamm import _compat  # noqa

_compat.register_backward_compatibility_import(
    __name__,
    "TransformerLayer",
    "tamm.layers.transformer.layer.TransformerLayer",
)



import torch as _torch

from tamm import _adapters_v1, _helpers
from tamm.layers import feed_forward as _feed_forward
from tamm.layers import side_outputs as _side_outputs
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class MixtureOfExperts(_torch.nn.Module, _LayerMixin):
    """
    A general mixture of experts layer.  This layer routes inputs, dispatches them,
    computes expert outputs, and then combines the results. The layer also supports
    optional norm (applied to the inputs), dropout (applied to the combined outputs),
    and residual connection (applied to the final outputs) transforms.

    Args:
        norm: A norm builder (optional).
        router: A :obj:`.MixtureOfExpertsRouter` builder.
        dispatcher: A :obj:`.MixtureOfExpertsDispatcher` builder.
        experts: A builder for the experts layer.  This layer must be compatible
            with the dispatcher (i.e., accept as inputs the ``experts_arg`` and
            ``experts_kwargs`` from the dispatcher, which depend on the
            dispatcher type).
        output_dropout: A :class:`.Dropout` builder (optional).
        residual_connection: A builder for a residual layer, such as
            :class:`.ResidualAdd` (optional).
        output_expert_assignments (:obj:`bool`): Flag that when ``True`` results in the
            inclusion of per token expert assignments as side outputs. Defaults to ``False``.
    """

    def __init__(
        self,
        norm: _OptionalModuleOrBuilder,
        router: _ModuleOrBuilder,
        dispatcher: _ModuleOrBuilder,
        experts: _ModuleOrBuilder,
        output_dropout: _OptionalModuleOrBuilder,
        residual_connection: _OptionalModuleOrBuilder,
        output_expert_assignments: bool = False,
    ):
        super().__init__()
        _helpers.append_children(
            self,
            norm=norm,
            router=router,
            dispatcher=dispatcher,
            experts=experts,
            output_dropout=output_dropout,
            residual_connection=residual_connection,
        )
        self._mark_adaptable_layers()
        self.output_expert_assignments = output_expert_assignments

    def _mark_adaptable_layers(self):
        if isinstance(self.experts, _feed_forward.TransformerFeedForward):
            _adapters_v1.layer_annotations.unannotate_layer(
                self.experts.output_transform
            )
            _adapters_v1.annotate_layer(
                self.experts.output_transform,
                [("MoEFeedForwardOutputTransform",)],
            )

            layers = (
                [
                    self.experts.hidden_transform.linear_0,
                    self.experts.hidden_transform.linear_1,
                ]
                if self.experts.activation.is_gated
                else [self.experts.hidden_transform]
            )

            for layer in layers:
                _adapters_v1.layer_annotations.unannotate_layer(layer)
                _adapters_v1.annotate_layer(
                    layer,
                    [("MoEFeedForwardHiddenTransform",)],
                )

    # pylint: disable-next=redefined-builtin
    def forward(self, input: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            input (:obj:`torch.Tensor`): Input hidden states, typically with shape
                ``(batch_size, seq_len, hidden_dim)`` for transformer models.

        Returns:
            A :obj:`torch.Tensor` or :obj:`OutputWithSideOutputs` (if the output
            contains auxiliary losses).  The main tensor output is the combined
            expert outputs.
        """

        x = input

        if self.norm is not None:
            x = self.norm(x)

        router_out = self.router(x)

        dispatch_out = self.dispatcher(
            x, mode="dispatch", expert_assignments=router_out.expert_assignments
        )

        x = self.experts(dispatch_out.experts_arg, **dispatch_out.experts_kwargs)

        x = self.dispatcher(
            x,
            mode="combine",
            assignment_weights=router_out.assignment_weights,
            combine_input=dispatch_out.combine_input,
        )

        if self.output_dropout is not None:
            x = self.output_dropout(x)

        if self.residual_connection is not None:
            x = self.residual_connection(x, residual_input=input)

        output = x
        side_outputs = {}

        if router_out.losses:
            side_outputs = _side_outputs.merge_side_outputs(
                side_outputs, other_side_outputs=router_out.losses
            )

        if self.output_expert_assignments:
            side_outputs = _side_outputs.merge_side_outputs(
                side_outputs,
                other_side_outputs={
                    "expert_assignments": router_out.expert_assignments
                },
            )

        if side_outputs:
            output = _side_outputs.OutputWithSideOutputs(
                output, side_outputs=side_outputs
            )

        return output
